# coding: utf8


def init_input_node(pet_nii):
    import nibabel as nib

    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_begin_image

    # Extract image ID
    image_id = get_subject_id(pet_nii)

    # Check that the PET file is a 3D volume
    img = nib.load(pet_nii)
    if len(img.shape) == 4:
        error_msg = f"Clinica does not handle 4D volumes for {image_id.replace('_', ' | ')}"
        cprint(error_msg, lvl="error")
        raise NotImplementedError(error_msg)

    # Print begin message
    print_begin_image(image_id)

    return pet_nii


def create_binary_mask(tissues, threshold=0.3):
    from os import getcwd
    from os.path import basename, join

    import nibabel as nib
    import numpy as np

    if len(tissues) == 0:
        raise RuntimeError(
            "The length of the list of tissues must be greater than zero."
        )

    img_0 = nib.load(tissues[0])
    shape = list(img_0.get_fdata(dtype="float32").shape)

    data = np.zeros(shape=shape)

    for image in tissues:
        data = data + nib.load(image).get_fdata(dtype="float32")

    data = (data > threshold) * 1.0
    out_mask = join(getcwd(), basename(tissues[0]) + "_brainmask.nii")

    mask = nib.Nifti1Image(data, img_0.affine, header=img_0.header)
    nib.save(mask, out_mask)
    return out_mask


def apply_binary_mask(image, binary_mask):
    from os import getcwd
    from os.path import basename, join

    import nibabel as nib

    original_image = nib.load(image)
    mask = nib.load(binary_mask)

    data = original_image.get_fdata(dtype="float32") * mask.get_fdata(dtype="float32")

    masked_image_path = join(getcwd(), "masked_" + basename(image))
    masked_image = nib.Nifti1Image(
        data, original_image.affine, header=original_image.header
    )
    nib.save(masked_image, masked_image_path)
    return masked_image_path


def create_pvc_mask(tissues):

    from os import getcwd
    from os.path import join

    import nibabel as nib
    import numpy as np

    if len(tissues) == 0:
        raise RuntimeError(
            "The length of the list of tissues must be greater than zero."
        )

    img_0 = nib.load(tissues[0])
    shape = img_0.get_fdata(dtype="float32").shape
    background = np.zeros(shape=shape)

    shape += tuple([len(tissues) + 1])
    data = np.empty(shape=shape, dtype=np.float64)

    for i in range(len(tissues)):
        image = nib.load(tissues[i])
        data[..., i] = np.array(image.get_fdata(dtype="float32"))
        background = background + image.get_fdata(dtype="float32")

    background = 1.0 - background
    data[..., len(tissues)] = np.array(background)

    out_mask = join(getcwd(), "pvc_mask.nii")
    mask = nib.Nifti1Image(data, img_0.affine, header=img_0.header)
    nib.save(mask, out_mask)
    return out_mask


def pet_pvc_name(pet_image, pvc_method):
    from os.path import basename

    pet_pvc_path = "pvc-" + pvc_method.lower() + "_" + basename(pet_image)
    return pet_pvc_path


def normalize_to_reference(pet_image, region_mask):
    from os import getcwd
    from os.path import basename, join

    import nibabel as nib
    import numpy as np

    pet = nib.load(pet_image)
    ref = nib.load(region_mask)

    region = np.multiply(pet.get_fdata(dtype="float32"), ref.get_fdata(dtype="float32"))
    region_mean = np.nanmean(np.where(region != 0, region, np.nan))

    data = pet.get_fdata(dtype="float32") / region_mean

    suvr_pet_path = join(getcwd(), "suvr_" + basename(pet_image))

    suvr_pet = nib.Nifti1Image(data, pet.affine, header=pet.header)
    nib.save(suvr_pet, suvr_pet_path)

    return suvr_pet_path


def atlas_statistics(in_image, in_atlas_list):
    """Generate regional measure from atlas_list in TSV files.

    For each atlas name provided it calculates for the input image the mean
    for each region in the atlas and saves it to a TSV file.

    Args:
        in_image: A Nifti image
        in_atlas_list: List of names of atlas to be applied

    Returns:
        List of paths to TSV files
    """
    from os import getcwd
    from os.path import abspath, join

    from nipype.utils.filemanip import split_filename

    from clinica.utils.atlas import AtlasAbstract
    from clinica.utils.statistics import statistics_on_atlas

    orig_dir, base, ext = split_filename(in_image)
    atlas_classes = AtlasAbstract.__subclasses__()
    atlas_statistics_list = []
    for atlas in in_atlas_list:
        for atlas_class in atlas_classes:
            if atlas_class.get_name_atlas() == atlas:
                out_atlas_statistics = abspath(
                    join(getcwd(), base + "_space-" + atlas + "_statistics.tsv")
                )
                statistics_on_atlas(in_image, atlas_class(), out_atlas_statistics)
                atlas_statistics_list.append(out_atlas_statistics)
                break

    return atlas_statistics_list


def get_from_list(in_list, index):
    return in_list[index]



from typing import Optional

import click

from clinica.pipelines import cli_param

pipeline_name = "pet-surface-longitudinal"


@click.command(name=pipeline_name)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.acq_label
@cli_param.argument.suvr_reference_region
@cli_param.argument.pvc_psf_tsv
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.n_procs
def cli(
    bids_directory: str,
    caps_directory: str,
    acq_label: str,
    suvr_reference_region: str,
    pvc_psf_tsv: str,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Longitudinal surface-based processing of PET images.

       ACQ_LABEL corresponds to the label given to the PET acquisition, specifying the tracer used. Frequently used values are 'fdg'
     or 'av45'.

       The reference region must be precised to perform intensity normalization. Accepted values include: 'pons', 'cerebellumPons', 'pons2', 'cerebellumPons2'.

       PVC_PSF_TSV is the TSV file containing the psf_x, psf_y and psf_z of the PSF for each PET image.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/PET_Surface_Longitudinal/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .pet_surface_pipeline import PetSurface

    parameters = {
        "acq_label": acq_label,
        "suvr_reference_region": suvr_reference_region,
        "pvc_psf_tsv": pvc_psf_tsv,
        "longitudinal": True,
    }

    pipeline = PetSurface(
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



# coding: utf8


def get_t1_freesurfer_custom_file():
    import os

    custom_file = os.path.join(
        "@subject",
        "@session",
        "t1",
        "freesurfer_cross_sectional",
        "@subject_@session",
        "surf",
        "@hemi.thickness.fwhm@fwhm.fsaverage.mgh",
    )
    return custom_file


def get_pet_surface_custom_file(acq_label, suvr_reference_region):
    import os

    custom_file = os.path.join(
        "@subject",
        "@session",
        "pet",
        "surface",
        f"@subject_@session_task-rest_acq-{acq_label}_pet"
        f"_space-fsaverage_suvr-{suvr_reference_region}_pvc-iy_hemi-@hemi_fwhm-@fwhm_projection.mgh",
    )
    return custom_file


def init_input_node(parameters, base_dir, subjects_visits_tsv):
    """Initialize the pipeline.

    This function will:
        - Create `surfstat_results_dir` in `base_dir`/<group_id> for SurfStat;
        - Save pipeline parameters in JSON file;
        - Copy TSV file with covariates;
        - Print begin execution message.
    """
    import json
    import os
    import shutil

    from clinica.pipelines.statistics_surface.statistics_surface_utils import (
        create_glm_info_dictionary,
    )
    from clinica.utils.ux import print_begin_image

    group_id = "group-" + parameters["group_label"]

    # Create surfstat_results_dir for SurfStat
    surfstat_results_dir = os.path.join(base_dir, group_id)
    os.makedirs(surfstat_results_dir, exist_ok=True)

    # Save pipeline parameters in JSON file
    glm_dict = create_glm_info_dictionary(subjects_visits_tsv, parameters)
    json_filename = os.path.join(surfstat_results_dir, group_id + "_glm.json")
    with open(json_filename, "w") as json_file:
        json.dump(glm_dict, json_file, indent=4)

    # Copy TSV file with covariates
    tsv_filename = os.path.join(surfstat_results_dir, group_id + "_covariates.tsv")
    shutil.copyfile(subjects_visits_tsv, tsv_filename)

    # Print begin message
    list_keys = ["AnalysisType", "Covariates", "Contrast", "FWHM", "ClusterThreshold"]
    list_values = [
        parameters["glm_type"],
        parameters["covariates"],
        parameters["contrast"],
        str(parameters["full_width_at_half_maximum"]),
        str(parameters["cluster_threshold"]),
    ]
    group_id = "group-" + parameters["group_label"]
    print_begin_image(group_id, list_keys, list_values)

    return parameters["group_label"], surfstat_results_dir


def get_string_format_from_tsv(tsv_file):
    """
    Determine string format from TSV file.

    If the TSV file is like:

    participant_id  session_id  sex     group   age
    sub-CLNC0001    ses-M00     Female  CN      71.1
    sub-CLNC0002    ses-M00     Male    CN      81.3
    sub-CLNC0003    ses-M00     Male    CN      75.4

    The columns of the TSV file contains consecutively strings, strings,
    strings, strings and float. The string_format is therefore "%s %s %s %s %f".

    Args:
        tsv_file: TSV file.

    Returns:
        String formatting of the TSV file (e.g. "%s %s %s %s %f")
    """
    import pandas as pd

    demographics_df = pd.read_csv(tsv_file, sep="\t")

    def dtype_to_str_format(dtype):
        """Convert pandas dtypes (e.g. int64) to string format (e.g. %d)"""
        import numpy as np

        if dtype == np.int64:
            str_format = "%d"
        elif dtype == np.float64:
            str_format = "%f"
        elif dtype == np.object:
            str_format = "%s"
        else:
            raise ValueError("Unknown dtype (given: %s)" % dtype)
        return str_format

    list_str_format = []
    for column in demographics_df.columns:
        list_str_format.append(dtype_to_str_format(demographics_df[column].dtype))

    return " ".join(list_str_format)


def covariates_to_design_matrix(contrast, covariates=None):
    """
    Generate design matrix for SurfStat based on the contrast and the optional list of covariates.

    Design matrix "1 + <contrast> + <covariate_1> + ... + <covariate_n>"

    Example:
        >>> from clinica.pipelines.statistics_surface.statistics_surface_utils import covariates_to_design_matrix
        >>> covariates_to_design_matrix('group', 'age sex group')
        1 + group + age + sex
        >>> covariates_to_design_matrix('group', 'age')
        1 + group + age
        >>> covariates_to_design_matrix('group', None)
        1 + group
    """
    if covariates:
        # Convert string to list while handling case where several spaces are present
        list_covariates = list(set(covariates.split(" ")))
        try:
            list_covariates.remove("")
        except ValueError:
            pass

        if contrast in list_covariates:
            design_matrix = "1 + " + " + ".join(
                covariate for covariate in list_covariates
            )
        else:
            design_matrix = (
                "1 + "
                + contrast
                + " + "
                + " + ".join(covariate for covariate in list_covariates)
            )
    else:
        design_matrix = "1 + " + contrast

    return design_matrix


def run_matlab(caps_dir, output_dir, subjects_visits_tsv, pipeline_parameters):
    """
    Wrap the call of SurfStat using clinicasurfstat.m Matlab script.

    Args:
        caps_dir (str): CAPS directory containing surface-based features
        output_dir (str): Output directory that will contain outputs of clinicasurfstat.m
        subjects_visits_tsv (str): TSV file containing the GLM information
        pipeline_parameters (dict): parameters of StatisticsSurface pipeline
    """
    import os

    from nipype.interfaces.matlab import MatlabCommand, get_matlab_command

    import clinica.pipelines as clinica_pipelines
    from clinica.pipelines.statistics_surface.statistics_surface_utils import (
        covariates_to_design_matrix,
        get_string_format_from_tsv,
    )
    from clinica.utils.check_dependency import check_environment_variable

    path_to_matlab_script = os.path.join(
        os.path.dirname(clinica_pipelines.__path__[0]), "lib", "clinicasurfstat"
    )
    freesurfer_home = check_environment_variable("FREESURFER_HOME", "FreeSurfer")

    MatlabCommand.set_default_matlab_cmd(get_matlab_command())
    matlab = MatlabCommand()
    matlab.inputs.paths = path_to_matlab_script
    matlab.inputs.script = """
    clinicasurfstat('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', %d, '%s', %.3f, '%s', %.3f, '%s', %.3f);
    """ % (
        os.path.join(caps_dir, "subjects"),
        output_dir,
        subjects_visits_tsv,
        covariates_to_design_matrix(
            pipeline_parameters["contrast"], pipeline_parameters["covariates"]
        ),
        pipeline_parameters["contrast"],
        get_string_format_from_tsv(subjects_visits_tsv),
        pipeline_parameters["glm_type"],
        pipeline_parameters["group_label"],
        freesurfer_home,
        pipeline_parameters["custom_file"],
        pipeline_parameters["measure_label"],
        "sizeoffwhm",
        pipeline_parameters["full_width_at_half_maximum"],
        "thresholduncorrectedpvalue",
        0.001,
        "thresholdcorrectedpvalue",
        0.05,
        "clusterthreshold",
        pipeline_parameters["cluster_threshold"],
    )
    # This will create a file: pyscript.m , the pyscript.m is the default name
    matlab.inputs.mfile = True
    # This will stop running with single thread
    matlab.inputs.single_comp_thread = False
    matlab.inputs.logfile = (
        "group-" + pipeline_parameters["group_label"] + "_matlab.log"
    )

    # cprint("Matlab logfile is located at the following path: %s" % matlab.inputs.logfile)
    # cprint("Matlab script command = %s" % matlab.inputs.script)
    # cprint("MatlabCommand inputs flag: single_comp_thread = %s" % matlab.inputs.single_comp_thread)
    # cprint("MatlabCommand choose which matlab to use(matlab_cmd): %s" % get_matlab_command())
    matlab.run()

    return output_dir


def create_glm_info_dictionary(tsv_file, pipeline_parameters):
    """Create dictionary containing the GLM information that will be stored in a JSON file."""
    out_dict = {
        # Clinica compulsory arguments
        "AnalysisType": pipeline_parameters["glm_type"],
        "DesignMatrix": covariates_to_design_matrix(
            pipeline_parameters["contrast"], pipeline_parameters["covariates"]
        ),
        "StringFormatTSV": get_string_format_from_tsv(tsv_file),
        "Contrast": pipeline_parameters["contrast"],
        "GroupLabel": pipeline_parameters["group_label"],
        # Optional arguments
        "Covariates": pipeline_parameters["covariates"],
        "FWHM": pipeline_parameters["full_width_at_half_maximum"],
        # Optional arguments for custom pipeline
        "custom_file": pipeline_parameters["custom_file"],
        "measure_label": pipeline_parameters["measure_label"],
        # Advanced arguments (i.e. tricky parameters)
        "ThresholdUncorrectedPvalue": 0.001,
        "ThresholdCorrectedPvalue": 0.05,
        "ClusterThreshold": pipeline_parameters["cluster_threshold"],
    }
    # Optional arguments for inputs from pet-surface pipeline
    if (
        pipeline_parameters["acq_label"]
        and pipeline_parameters["suvr_reference_region"]
    ):
        out_dict["acq_label"] = pipeline_parameters["acq_label"]
        out_dict["suvr_reference_region"] = pipeline_parameters["suvr_reference_region"]

    return out_dict


def save_to_caps(source_dir, caps_dir, overwrite_caps, pipeline_parameters):
    """Save `source_dir`/ to CAPS folder.

    This function copies outputs of `source_dir`/ to
    `caps_dir`/groups/<group_id>/<statistics>/surfstat_<glm_type>/

    The `source_dir`/ folder should contain the following elements:
        - group-<group_label>_<group_1_or_2>-lt-<group_1_or_2>_measure-<measure>_fwhm-<label>_suffix.ext
    or
        - group-<group_label>_correlation-<label>_contrast-{-|+}_measure-<measure>_fwhm-<label>_suffix.ext
    and
        - group-<group_label>_covariates.tsv
        - group-<group_label>_glm.json

    Raise:
        NotImplementedError: If overwrite_caps=True.
    """
    import os
    import shutil

    from clinica.utils.ux import print_end_image

    group_id = "group-" + pipeline_parameters["group_label"]

    if pipeline_parameters["glm_type"] == "group_comparison":
        surfstat_folder = "surfstat_" + pipeline_parameters["glm_type"]
    elif pipeline_parameters["glm_type"] == "correlation":
        surfstat_folder = "surfstat_" + pipeline_parameters["glm_type"] + "_analysis"
    else:
        raise NotImplementedError(
            "The other GLM situations have not been implemented in this pipeline."
        )

    destination_dir = os.path.join(
        os.path.expanduser(caps_dir), "groups", group_id, "statistics", surfstat_folder
    )

    if overwrite_caps:
        raise NotImplementedError("save_to_caps(overwrite_caps=True) not implemented")
    shutil.copytree(source_dir, destination_dir, symlinks=True)
    print_end_image(group_id)



# coding: utf8

"""Utils to convert AIBL dataset in BIDS."""


def listdir_nohidden(path):
    """List all the subdirectories of path except the hidden folders.

    Args:
        path (str): path whose subdirectories are needed

    Returns:
        List(str): list of all the subdirectories of path
    """
    from os import listdir

    return [result for result in listdir(path) if not result.startswith(".")]


def find_T1_folder(subdirectory, path_to_T1_1):
    """Check if the subdirectory contains a T1 image, and it return the h.

    Args:
        subdirectory (str): subdirectory: name of the folder
        path_to_T1_1 (str): [description]

    Returns:
        str: previous path to arrive to the T1 image
    """
    import os

    path_to_convert = {
        "MPRAGE_ADNI_confirmed",
        "MPRAGE",
        "MPRAGE_ADNI_confirmed_RPT",
        "MPRAGE_ADNI_confirmed_REPEATX2",
        "MPRAGE_ADNI_confirmed_repeat",
        "MPRAGE_ADNI_confirmed_REPEAT",
        "MPRAGE_ADNI_conf_REPEAT",
    }
    for j in path_to_convert:
        path = []
        # if conditions which checks if the subfolder contain a T1 image
        if j == subdirectory:
            path = os.path.join(path_to_T1_1, subdirectory)
            return path
    if path == []:
        return "NaN"  # there are no more folders which could contain T1 images


def find_T1_folder_nodata(subdirectory, path_to_T1_1):
    """

    This method checks if the subdirectory contains a T1 image, and it
    returns the path. This method differs from the find_T1_folder since for
    these folders the exame_date is not present in the clinical excel file
    and we will not check if the exame_date corresponds to the date stored
    in the path to the image, but they will be converted anyway

     :param subdirectory: name of the folder

     :return: previous path to arrive to the T1 image
    """
    import os

    path_to_convert = {"MPRAGESAGISOp2ND", "MPRAGE_SAG_ISO_p2_ND", "MPRAGE_SAG_ISO_p2"}
    for j in path_to_convert:
        path = []
        # if conditions which checks if the subfolder contain a T1 image
        if j == subdirectory:
            path = os.path.join(path_to_T1_1, subdirectory)
            return path
    if path == []:
        return "NaN"  # there are no more folders which could contain T1 images


def find_correspondance_index(i, csv_file):
    """Give as output the index of the CSV file analysed which correspond to the 'i' subject.

    :param i: subject_ID
    :param csv_file: CSV file where all the information are listed

    :return: index
    """
    index = []
    for x in csv_file.RID:
        if i == str(x):
            index = csv_file.RID[csv_file.RID == x].index.tolist()
            return index


def find_correspondance_date(index, csv_file):
    """Return the dates reported in the csv_file for the i-subject.

    :param index: index corresponding to the subject analysed
    :param csv_file: csv file where all the information are listed
    :return date
    """
    return csv_file.EXAMDATE[index]


def match_data(exame_date, i, csv_file):
    """

    This method returns the session_ID. It controls if the dates
    corresponding to the image (from the name of the subdirectory)
    correspond to one of the dates listed from the csv_file for the subject
    analysed. The session_ID is the corresponding session for that patient
    in that date.  It returns -4 if there are no information.

    :param exame_date: date where the image has been taken, it is saved
    from the name of the corresponding subdirector
    :param i: subject_ID
    :param csv_file: csv file where all the information are listed

    :return session_id of the patient
    """
    import re

    session_ID = []
    index = find_correspondance_index(i, csv_file)
    csv_date = find_correspondance_date(index, csv_file)
    for xx in index:
        if str(csv_date[xx]) != "-4":
            # check is the date is not '-4'
            m = re.search(
                "([0-9].*)-(.*)-(.*)_(.*)_(.*)_(.*)", str(exame_date)
            )  # string from image directory
            p = re.search(
                "(.*)/(.*)/(.*)", str(csv_date[xx])
            )  # string from the date of the csv_file
            if (
                (p.group(1) == m.group(2))
                & (p.group(2) == m.group(3))
                & (p.group(3) == m.group(1))
            ):
                session_ID = csv_file.VISCODE[xx]
    if session_ID == []:
        session_ID = "-4"
    return session_ID


def list_of_paths():
    """List all the folders which not contain PET images."""
    return [
        ".DS_Store",
        "localizer",
        "Space_3D_T2_FLAIR_sag_p2",
        "AXIAL_FLAIR",
        "MPRAGE_ADNI_confirmed_REPEATX2",
        "Axial_PD-T2_TSE",
        "Axial_PD-T2_TSE_repeat",
        "MPRAGE_SAG_ISO_p2_ND",
        "Axial_PD-T2_TSE_confirmed",
        "MPRAGESAGISOp2ND",
        "MPRAGE_ADNI_confirmed",
        "MPRAGE_ADNI_confirmed_repeat",
        "MPRAGE_SAG_ISO_p2",
        "MPRAGE",
        "MPRAGE_ADNI_confirmed_REPEAT",
        "Axial_PD-T2_TSE_confirmed_repeat",
        "MPRAGE_ADNI_conf_REPEAT",
        "Space_3D_T2_FLAIR_sag_p2_REPEAT",
        "MPRAGE_ADNI_confirmed_RPT",
        "Brain_256_1.6_zoom_4_x_4_iter",
        "Space_3D_T2_FLAIR_sag_REPEAT",
        "Axial_PD-T2_TSE_RPTconfirmed",
        "Axial_PD-T2_TSE_RPT_confirmed",
        "Axial_PD-T2_TSE_confirmed_REPEAT",
        "flair_t2_spc_irprep_ns_sag_p2_1mm_iso",
        "localiser",
    ]


def check_subdirectories_pet(subdirectories, sub, no_pet):
    """

    It returns the correct subdirectories for the PET images, they should
    belong to the list where there all the possible names of the PET images.

    :param subdirectories:
    :param sub: all the possible subdirectories which need to be checked
    :param no pet: list of names of folders which not contain PET images

    :return subdirectory which is containing a PET image which needs to be
    converted
    """

    for j in range(len(sub)):
        if (sub[j] not in no_pet) & (sub[j] != ".DS_Store"):
            subdirectories.append(sub[j])
    subdirectories = list(set(subdirectories))
    return subdirectories


def dicom_to_nii(subject, output_path, output_filename, image_path):
    """Convert the dicom images in a nifti files using dicom2nii or mri_convert.

    :param subject:
    :param output_path: where nifti image is stored
    :param output_filename: name of the nifti image
    :param image_path: where dicom files are stored

    :return: Image in a nifti format
    """
    import os
    import shutil
    import subprocess
    from os.path import exists

    from clinica.iotools.bids_utils import run_dcm2niix
    from clinica.utils.stream import cprint

    try:
        os.makedirs(output_path)
    except OSError:
        if not os.path.isdir(output_path):
            raise

    # if image.Is_Dicom:
    command = f"dcm2niix -b n -z y -o {output_path} -f {output_filename} {image_path}"
    run_dcm2niix(command)

    nifti_file = os.path.join(output_path, output_filename + ".nii.gz")

    if not exists(nifti_file):
        cprint(nifti_file + " should have been created but this did not happen")
    return nifti_file


def viscode_to_session(viscode):
    """Replace the session label 'bl' with 'M00' or capitalize the session name passed as input.

    :param viscode: session name

    :return: M00 if is the baseline session or the original session name
    capitalized
    """
    if viscode == "bl":
        return "M00"
    else:
        return viscode.capitalize()


def find_path_to_pet_modality(path_to_dataset, csv_file):
    """Create a Dataframe which contains all the paths to the PET image of a modality (for example AV45 or PIB).

    :param path_to_dataset: path to AIBL dataset
    :param csv_file: file which correspond to the modality

    :return: A dataframe which contains the path for PET images for a
    single modality and subject_ID and session_ID are reported for each
    path
    """
    import os

    import pandas

    # TODO
    # exclude_subjects = get_exclude_subject(file.txt)

    no_pet = list_of_paths()
    subjects_ID = listdir_nohidden(path_to_dataset)
    # selection of the subjects_ID from the folder downloaded
    # this subject must be discarded since it is only a sample and not a patient
    if "0151083" in subjects_ID:
        del subjects_ID[subjects_ID.index("0151083")]
    sub_ID = []
    ses_ID = []
    path_pet = []

    # Iteration through all the subjects_ID
    def is_int(x):
        for i in x:
            if int(i) in list(csv_file.RID):
                yield i

    #    def append_path(image_ID):

    for i in is_int(subjects_ID):
        # check if the subject is present in the csv_file for the modality selected
        subdirectories = []
        path_to_pet_1 = os.path.join(path_to_dataset, str(i))
        # subdirectory_all = os.listdir(path_to_pet_1)
        subdirectory_all = listdir_nohidden(path_to_pet_1)
        subdirectories = check_subdirectories_pet(
            subdirectories, subdirectory_all, no_pet
        )
        # selection only of the folders which contain PET image
        for j in range(len(subdirectories)):
            path_to_pet_2 = os.path.join(path_to_pet_1, subdirectories[j])
            exame_date = listdir_nohidden(path_to_pet_2)
            # exame date of the image which is going to be converted
            for x in range(len(exame_date)):
                # selection of the session_ID matching the data in the csv_file with the one of the image
                session_ID = match_data(exame_date[x], i, csv_file)
                if session_ID != "-4":
                    path_to_pet_3 = os.path.join(path_to_pet_2, str(exame_date[x]))
                    # For the RID 1607 there are two PET images of the flute modality, and we select the first
                    if i == "1607":
                        if subdirectories[j] == "Flute_256_1.6_Zoom_plain_4_x_4_Iter":
                            image_ID = ["I442930"]
                        else:
                            image_ID = listdir_nohidden(path_to_pet_3)
                    else:
                        image_ID = listdir_nohidden(path_to_pet_3)

                    for y in range(len(image_ID)):
                        # final path to find the image we want to convert
                        path_to_pet = os.path.join(path_to_pet_3, image_ID[y])  #
                        sub_ID.append(i)
                        ses_ID.append(session_ID)
                        path_pet.append(path_to_pet)

    data = pandas.DataFrame(
        {"Subjects_ID": sub_ID, "Session_ID": ses_ID, "Path_to_pet": path_pet}
    )
    # data=final dataframe

    return data


def find_path_to_T1_ADNI(file_mri, subjects_ID, path_to_dataset):
    """

    This method creates a Dataframe which contains all the paths to the T1
    images which are ADNI compliant (as explained in the AIBL website).
    This images differ from the others T1 of the dataset since in the
    cvs_file is reported the exame date.

    :param file_mri: in the clinical data there are two files which
    describe the  parameters of the T1 images (MRI 1.5 T and MRI 3T)
    :param subjects_ID: subjects_id in the dataset dowloaded
    :param path_to_dataset: path to AIBL dataset

    :return: A dataframe which contains the path for T1 images and
    subject_ID and session_ID are reported for each path
    """
    import os

    sub_ID = []
    ses_ID = []
    path_T1 = []

    for i in subjects_ID:
        for jj in file_mri:
            # it checks all the file_mri
            if int(i) in list(jj.RID):
                # check if the information of the subject are present in the csv_file
                path_to_T1_1 = os.path.join(path_to_dataset, str(i))
                # subdirectories = os.listdir(path_to_T1_1)
                subdirectories = listdir_nohidden(path_to_T1_1)
                for j in range(len(subdirectories)):
                    # check if the subdirectory can contain a T1 image
                    path_to_T1_2 = find_T1_folder(subdirectories[j], path_to_T1_1)
                    if path_to_T1_2 != "NaN":
                        exame_date = listdir_nohidden(
                            path_to_T1_2
                        )  # this is the string I need to compare with the csv
                        for x in range(len(exame_date)):
                            # check if the corresponding session_ID can be found in the csv_file
                            session_ID = match_data(exame_date[x], i, jj)
                            if session_ID != "-4":
                                path_to_T1_3 = os.path.join(
                                    path_to_T1_2, str(exame_date[x])
                                )
                                image_ID = listdir_nohidden(path_to_T1_3)
                                for y in range(len(image_ID)):
                                    # compute the final path
                                    path_to_T1 = os.path.join(path_to_T1_3, image_ID[y])
                                    sub_ID.append(i)
                                    ses_ID.append(session_ID)
                                    path_T1.append(path_to_T1)

    return [sub_ID, ses_ID, path_T1]


def find_path_to_T1_SAG(path_to_dataset, subjects_ID, sub_ID, ses_ID, path_T1):
    """

    This method creates a Dataframe which contains all the paths to the T1
    images which are not ADNI compliant, they contain the word "SAG" in
    their name

    :param path_to_dataset: path to AIBL dataset
    :param subjects_ID: subjects_id in the dataset dowloaded
    :param sub_ID: the previous list (from T1_ADNI) where new subjects ID
    will be appended
    :param ses_ID: the previous list (from T1_ADNI) where new session ID
    will be appended
    :param path_T1:the previous list (from T1_ADNI) where new paths will be
    appended

    :return: it completes the list of all the T1 paths including all the
    images where we didn't find the exame-data but we can fix it with a
    further analysis
    """
    import os

    for i in subjects_ID:
        subdirectory_for_subject = []
        path_to_T1_1 = os.path.join(path_to_dataset, str(i))
        # subdirectories = os.listdir(path_to_T1_1)
        subdirectories = listdir_nohidden(path_to_T1_1)
        for j in range(len(subdirectories)):
            # we convert only the images which are in this list and we take only one of them for subject
            if subdirectories[j] in [
                "MPRAGESAGISOp2ND",
                "MPRAGE_SAG_ISO_p2_ND",
                "MPRAGE_SAG_ISO_p2",
            ]:
                subdirectory_for_subject.append(subdirectories[j])
        if not subdirectory_for_subject:
            pass
        else:
            path_to_T1_2 = os.path.join(path_to_T1_1, subdirectory_for_subject[0])

            exame_date = listdir_nohidden(path_to_T1_2)
            if i in [342, 557]:
                session_ID = "M54"
            else:
                session_ID = "M00"
            if (i in sub_ID and session_ID != ses_ID[sub_ID.index(i)]) or (
                i not in sub_ID
            ):
                # if for a subject in the same session we have both this image and the "ADNI" compliant we are converting the second one since the exame-date is more precise
                path_to_T1_3 = os.path.join(path_to_T1_2, str(exame_date[0]))
                image_ID = listdir_nohidden(path_to_T1_3)
                path_to_T1 = os.path.join(path_to_T1_3, image_ID[0])
                # we append the result to the list
                sub_ID.append(i)
                ses_ID.append(session_ID)
                path_T1.append(path_to_T1)

    return [sub_ID, ses_ID, path_T1]


def find_path_to_T1(path_to_dataset, path_to_csv):
    """
    This method creates a DataFrame for the T1 images, where for each of
    them the subject ID, the session ID and the path to the image are
    reported

    :param path_to_dataset:  path to AIBL dataset
    :param path_to_csv: path to the csv files downloaded
    :return: pandas dataframe which contains all the paths for the T1
    images, and the correisponding subject_ID and session_ID
    """
    import glob
    import os

    import pandas

    # two csv_files contain information regarding the T1w MRI images
    mri_meta = pandas.read_csv(
        glob.glob(os.path.join(path_to_csv, "aibl_mrimeta_*.csv"))[0]
    )
    mri_3meta = pandas.read_csv(
        glob.glob(os.path.join(path_to_csv, "aibl_mri3meta_*.csv"))[0]
    )
    file_mri = [mri_meta, mri_3meta]
    subjects_ID = listdir_nohidden(path_to_dataset)
    # list of all the folders which correspond to the subject_ID
    # all the subjects downloaded are taken into account for the conversion, except this sample
    if "0151083" in subjects_ID:
        del subjects_ID[subjects_ID.index("0151083")]
    [sub_ID, ses_ID, path_T1] = find_path_to_T1_ADNI(
        file_mri, subjects_ID, path_to_dataset
    )
    [sub_ID, ses_ID, path_T1] = find_path_to_T1_SAG(
        path_to_dataset, subjects_ID, sub_ID, ses_ID, path_T1
    )

    data = pandas.DataFrame(
        {"Subjects_ID": sub_ID, "Session_ID": ses_ID, "Path_to_T1": path_T1}
    )
    # data= final dataframe
    return data


# Covert the AIBL PET images into the BIDS specification.
# There are three pet modalities: av45, pib, flute. All of them are converted
# in BIDS


def paths_to_bids(path_to_dataset, path_to_csv, bids_dir, modality, overwrite=False):
    """Convert all the T1 images found in the AIBL dataset downloaded in BIDS.

    Args:
        path_to_dataset: path_to_dataset
        path_to_csv: path to the csv file containing clinical data
        bids_dir: path to save the AIBL-T1-dataset converted in a BIDS format
        modality: string 't1', 'av45', 'flute' or 'pib'
        overwrite: if True previous existing outputs will be erased

    Returns:
        list of all the images that are potentially converted in a BIDS format and saved in the bids_dir.
        This does not guarantee existence.
    """
    import glob
    from multiprocessing import Value, cpu_count
    from multiprocessing.dummy import Pool
    from os import makedirs, remove
    from os.path import exists, join

    import pandas as pds
    from numpy import nan

    from clinica.iotools.bids_utils import json_from_dcm
    from clinica.iotools.utils.data_handling import center_nifti_origin
    from clinica.utils.stream import cprint

    if modality.lower() not in ["t1", "av45", "flute", "pib"]:
        # This should never be reached
        raise RuntimeError(f"{modality.lower()} is not supported for conversion")

    counter = None

    makedirs(join(bids_dir, "conversion_info"), exist_ok=True)

    def init(args):
        """Store the counter for later use."""
        global counter
        counter = args

    def create_file(image):
        global counter
        subject = image.Subjects_ID
        session = image.Session_ID
        name_of_path = {
            "t1": "Path_to_T1",
            "av45": "Path_to_pet",
            "flute": "Path_to_pet",
            "pib": "Path_to_pet",
        }
        # depending on the DataFrame, there is different way of accessing
        # the image object
        image_path = image[name_of_path[modality]]
        with counter.get_lock():
            counter.value += 1
            if image_path is nan:
                cprint(f"No path specified for {subject} in session {session}")
            else:
                cprint(
                    f"'[{modality.upper()}] Processing subject {subject} - {session} {session}, {counter.value}/{total}"
                )

        if image_path is nan:
            return nan
        session = viscode_to_session(session)
        # creation of the path
        if modality == "t1":
            output_path = join(bids_dir, f"sub-AIBL{subject}", f"ses-{session}", "anat")
            output_filename = f"sub-AIBL{subject}_ses-{session}_T1w"
        elif modality in ["flute", "pib", "av45"]:
            output_path = join(bids_dir, f"sub-AIBL{subject}", f"ses-{session}", "pet")
            output_filename = (
                f"sub-AIBL{subject}_ses-{session}_task-rest_acq-{modality}_pet"
            )
        # image is saved following BIDS specifications

        if exists(join(output_path, output_filename + ".nii.gz")) and not overwrite:
            cprint(f"Subject {str(subject)} - session {session} already processed.")
            output_image = join(output_path, output_filename + ".nii.gz")
        else:
            if exists(join(output_path, output_filename + ".nii.gz")):
                remove(join(output_path, output_filename + ".nii.gz"))
            output_image = dicom_to_nii(
                subject, output_path, output_filename, image_path
            )
            json_from_dcm(image_path, join(output_path, output_filename + ".json"))

        # Center all images
        center_nifti_origin(output_image, output_image)

        return output_image

    # it reads the DataFrame where subject_ID, session_ID and path are saved
    if modality == "t1":
        images = find_path_to_T1(path_to_dataset, path_to_csv)
    else:
        path_to_csv_pet_modality = glob.glob(
            join(path_to_csv, "aibl_" + modality + "meta_*.csv")
        )[0]
        if not exists(path_to_csv_pet_modality):
            raise FileNotFoundError(
                path_to_csv_pet_modality + " file not found in clinical data folder"
            )
        # Latest version of Flutemetamol CSV file (aibl_flutemeta_01-Jun-2018.csv)
        # has an extra column for some rows. However, each CSV file (regarding PET tracers)
        # contains the same columns. The usecols fixes this issue.
        df_pet = pds.read_csv(
            path_to_csv_pet_modality, sep=",|;", usecols=list(range(0, 36))
        )
        images = find_path_to_pet_modality(path_to_dataset, df_pet)
    images.to_csv(
        join(bids_dir, "conversion_info", modality + "_paths.tsv"),
        index=False,
        sep="\t",
        encoding="utf-8",
    )

    counter = Value("i", 0)
    total = images.shape[0]
    # Reshape inputs to give it as a list to the workers
    images_list = []
    for i in range(total):
        images_list.append(images.iloc[i])

    # intializer are used with the counter variable to keep track of how many
    # files have been processed
    poolrunner = Pool(cpu_count(), initializer=init, initargs=(counter,))
    output_file_treated = poolrunner.map(create_file, images_list)
    del counter
    return output_file_treated


# -- Methods for the clinical data --


def create_participants_df_AIBL(
    input_path, clinical_spec_path, clinical_data_dir, delete_non_bids_info=True
):
    """Create a participants file for the AIBL dataset where information regarding the patients are reported.

    Args:
        input_path: path to the input directory
        clinical_spec_path: path to the clinical file
        clinical_data_dir: directory to the clinical data files
        delete_non_bids_info: if True delete all the rows of the subjects
        that are not available in the BIDS dataset

    Returns:
        a pandas DataFrame that contains the participants data
    """
    import glob
    import os
    import re
    from os import path

    import numpy as np
    import pandas as pd

    fields_bids = ["participant_id"]
    fields_dataset = []
    prev_location = ""
    index_to_drop = []

    location_name = "AIBL location"
    clinical_spec_path = clinical_spec_path + "_participant.tsv"

    if not os.path.exists(clinical_spec_path):
        raise FileNotFoundError(clinical_spec_path + " not found in clinical data.")
    participants_specs = pd.read_csv(clinical_spec_path, sep="\t")
    participant_fields_db = participants_specs["AIBL"]
    field_location = participants_specs[location_name]
    participant_fields_bids = participants_specs["BIDS CLINICA"]

    # Extract the list of the available fields for the dataset (and the corresponding BIDS version)
    for i in range(0, len(participant_fields_db)):
        if not pd.isnull(participant_fields_db[i]):
            fields_bids.append(participant_fields_bids[i])
            fields_dataset.append(participant_fields_db[i])

    # Init the dataframe that will be saved in the file participant.tsv
    participant_df = pd.DataFrame(columns=fields_bids)

    csv_files = []
    for i in range(0, len(participant_fields_db)):
        # If a field not empty is found
        if not pd.isnull(participant_fields_db[i]):
            # Extract the file location of the field and read the value from the file
            tmp = field_location[i].split("/")
            location = tmp[0]
            # If a sheet is available
            if len(tmp) > 1:
                sheet = tmp[1]
            else:
                sheet = ""
            # Check if the file to open for a certain field it's the same of the previous field
            if location == prev_location and sheet == prev_sheet:
                pass
            else:
                file_ext = os.path.splitext(location)[1]
                file_to_read_path = path.join(clinical_data_dir, location)

                if file_ext == ".xlsx":
                    file_to_read = pd.read_excel(
                        glob.glob(file_to_read_path)[0], sheet_name=sheet
                    )
                elif file_ext == ".csv":
                    file_to_read = pd.read_csv(glob.glob(file_to_read_path)[0])
                prev_location = location
                prev_sheet = sheet

            field_col_values = []
            # For each field in fields_dataset extract all the column values
            for j in range(0, len(file_to_read)):
                # Convert the alternative_id_1 to string if is an integer/float
                if participant_fields_bids[i] == "alternative_id_1" and (
                    file_to_read[participant_fields_db[i]].dtype == np.float64
                    or file_to_read[participant_fields_db[i]].dtype == np.int64
                ):
                    if not pd.isnull(file_to_read.at[j, participant_fields_db[i]]):
                        # value_to_append = str(file_to_read.get_value(j, participant_fields_db[i])).rstrip('.0')
                        value_to_append = str(
                            file_to_read.at[j, participant_fields_db[i]]
                        )

                    else:
                        value_to_append = "n/a"
                else:
                    value_to_append = file_to_read.at[j, participant_fields_db[i]]
                field_col_values.append(value_to_append)
            # Add the extracted column to the participant_df
            participant_df[participant_fields_bids[i]] = pd.Series(field_col_values)

    # Adding participant_id column with BIDS ids
    for i in range(0, len(participant_df)):
        value = participant_df["alternative_id_1"][i]
        participant_df["participant_id"][i] = "sub-AIBL" + str(value)
        participant_df["date_of_birth"][i] = re.search(
            "/([0-9].*)", str(participant_df["date_of_birth"][i])
        ).group(1)
        if participant_df["sex"][i] == 1:
            participant_df["sex"][i] = "M"
        else:
            participant_df["sex"][i] = "F"

    participant_df.replace("-4", "n/a")

    # Delete all the rows of the subjects that are not available in the BIDS dataset
    if delete_non_bids_info:
        participant_df = participant_df.drop(index_to_drop)

    participant_df.to_csv(
        os.path.join(input_path, "participants.tsv"),
        sep="\t",
        index=False,
        encoding="utf8",
    )

    return participant_df


def create_sessions_dict_AIBL(input_path, clinical_data_dir, clinical_spec_path):
    """Extract the information regarding the sessions and store them in a dictionary.

    :param input_path: path to the input folder
    :param clinical_spec_path: path to the clinical file
    :param clinical_data_dir: directory to the clinical data files
    :return: A dataframe saved in a tsv file which contains information for each session
    """
    import glob
    from os import path

    import pandas as pd

    # Load data
    location = "AIBL location"
    sessions = pd.read_csv(clinical_spec_path + "_sessions.tsv", sep="\t")
    sessions_fields = sessions["AIBL"]
    field_location = sessions[location]
    sessions_fields_bids = sessions["BIDS CLINICA"]
    fields_dataset = []
    fields_bids = []
    sessions_dict = {}

    for i in range(0, len(sessions_fields)):
        if not pd.isnull(sessions_fields[i]):
            fields_bids.append(sessions_fields_bids[i])
            fields_dataset.append(sessions_fields[i])

    sessions_df = pd.DataFrame(columns=fields_bids)

    files_to_read = []
    sessions_fields_to_read = []
    for i in range(0, len(sessions_fields)):
        # If the i-th field is available
        if not pd.isnull(sessions_fields[i]):
            # Load the file
            tmp = field_location[i]
            location = tmp[0]
            file_to_read_path = path.join(clinical_data_dir, tmp)
            files_to_read.append(glob.glob(file_to_read_path)[0])
            sessions_fields_to_read.append(sessions_fields[i])

    rid = pd.read_csv(files_to_read[0], dtype={"text": str}, low_memory=False).RID
    rid = list(set(rid))
    for r in rid:
        dict = []
        for i in files_to_read:
            file_to_read = pd.read_csv(i, dtype={"text": str})
            if len(file_to_read.columns) == 1:
                file_to_read = pd.read_csv(i, sep=";", low_memory=False)

            # information are written following the BIDS specifications
            viscode = file_to_read.loc[(file_to_read["RID"] == r), "VISCODE"]
            for j in sessions_fields_to_read:
                if j in list(file_to_read.columns.values) and j == "MMSCORE":
                    MMSCORE = file_to_read.loc[(file_to_read["RID"] == r), j]
                    MMSCORE[MMSCORE == -4] = "n/a"
                elif j in list(file_to_read.columns.values) and j == "CDGLOBAL":
                    CDGLOBAL = file_to_read.loc[(file_to_read["RID"] == r), j]
                    CDGLOBAL[CDGLOBAL == -4] = "n/a"
                elif j in list(file_to_read.columns.values) and j == "DXCURREN":
                    DXCURREN = file_to_read.loc[(file_to_read["RID"] == r), j]
                    DXCURREN[DXCURREN == -4] = "n/a"
                    DXCURREN[DXCURREN == 1] = "CN"
                    DXCURREN[DXCURREN == 2] = "MCI"
                    DXCURREN[DXCURREN == 3] = "AD"
                elif j in list(file_to_read.columns.values) and j == "EXAMDATE":
                    EXAMDATE = file_to_read.loc[(file_to_read["RID"] == r), j]
                elif j in list(file_to_read.columns.values) and j == "PTDOB":
                    PTDOB = file_to_read.loc[(file_to_read["RID"] == r), j]

        examdates = get_examdates(
            r, EXAMDATE.to_list(), viscode.to_list(), clinical_data_dir
        )
        age = get_ages(PTDOB.values[0], examdates)

        viscode[viscode == "bl"] = "M00"
        viscode = viscode.str.upper()

        dict = pd.DataFrame(
            {
                "session_id": "ses-" + viscode,
                "age": age,
                "MMS": MMSCORE,
                "cdr_global": CDGLOBAL,
                "diagnosis": DXCURREN,
                "examination_date": examdates,
            }
        )

        cols = dict.columns.tolist()
        dict = dict[cols[-1:] + cols[:-1]]

        bids_paths = path.join(input_path, "sub-AIBL" + str(r))
        if path.exists(bids_paths):
            dict.to_csv(
                path.join(
                    input_path,
                    "sub-AIBL" + str(r),
                    "sub-AIBL" + str(r) + "_sessions.tsv",
                ),
                sep="\t",
                index=False,
                encoding="utf8",
            )


def create_scans_dict_AIBL(input_path, clinical_data_dir, clinical_spec_path):
    """Create scans.tsv files for AIBL.

    Args:
        input_path: path to the input folder
        clinical_spec_path: path to the clinical file
        clinical_data_dir: directory to the clinical data files
    """
    import glob
    from os import path

    import pandas as pd

    import clinica.iotools.bids_utils as bids

    # Load data
    location = "AIBL location"
    scans = pd.read_csv(clinical_spec_path + "_scans.tsv", sep="\t")
    scans_fields = scans["AIBL"]
    field_location = scans[location]
    scans_fields_bids = scans["BIDS CLINICA"]
    fields_dataset = []
    fields_bids = []

    # Keep only fields for which there are AIBL fields
    for i in range(0, len(scans_fields)):
        if not pd.isnull(scans_fields[i]):
            fields_bids.append(scans_fields_bids[i])
            fields_dataset.append(scans_fields[i])

    files_to_read = []
    sessions_fields_to_read = []
    for i in range(0, len(scans_fields)):
        # If the i-th field is available
        if not pd.isnull(scans_fields[i]):
            # Load the file
            tmp = field_location[i]
            file_to_read_path = path.join(clinical_data_dir, tmp)
            files_to_read.append(glob.glob(file_to_read_path)[0])
            sessions_fields_to_read.append(scans_fields[i])

    bids_ids = [
        path.basename(sub_path)
        for sub_path in glob.glob(path.join(input_path, "sub-AIBL*"))
    ]

    # This dictionary should be automatically computed from the dataset
    #
    ses_dict = bids.get_sessions_map_AIBL(bids_ids, input_path)

    scans_dict = bids.create_scans_dict(
        clinical_data_dir,
        "AIBL",
        clinical_spec_path,
        bids_ids,
        "RID",
        "VISCODE",
        ses_dict,
    )
    bids.write_scans_tsv(input_path, bids_ids, scans_dict)


def get_examdates(rid, examdates, viscodes, clinical_data_dir):
    import glob
    from datetime import datetime
    from os import path

    import pandas as pd
    from dateutil.relativedelta import relativedelta

    res_examdates = []
    csv_list = [
        glob.glob(path.join(clinical_data_dir, "aibl_mri3meta_*.csv"))[0],
        glob.glob(path.join(clinical_data_dir, "aibl_mrimeta_*.csv"))[0],
        glob.glob(path.join(clinical_data_dir, "aibl_cdr_*.csv"))[0],
        glob.glob(path.join(clinical_data_dir, "aibl_flutemeta_*.csv"))[0],
        glob.glob(path.join(clinical_data_dir, "aibl_mmse_*.csv"))[0],
        glob.glob(path.join(clinical_data_dir, "aibl_pibmeta_*.csv"))[0],
    ]

    for e in range(len(examdates)):
        exam = examdates[e]

        if exam != "-4":
            res_examdates.append(exam)
            continue

        # If EXAMDATE does not exist (-4) we try to obtain it from another .csv file
        for csv_file in csv_list:
            if "aibl_flutemeta" in csv_file:
                csv_data = pd.read_csv(
                    csv_file, low_memory=False, usecols=list(range(0, 36))
                )
            else:
                csv_data = pd.read_csv(csv_file, low_memory=False)
            exam_date = csv_data[
                (csv_data.RID == rid) & (csv_data.VISCODE == viscodes[e])
            ]
            if not exam_date.empty and exam_date.iloc[0].EXAMDATE != "-4":
                exam = exam_date.iloc[0].EXAMDATE
                break

        # If EXAMDATE still does not exist (-4) we add the session months to baseline date
        if exam == "-4":
            bl_index = viscodes.index("bl")
            if bl_index > -1:
                bl_date = examdates[bl_index]
                bl_examdate = datetime.strptime(bl_date, "%m/%d/%Y")
                if viscodes[e] != "bl":
                    months = int(viscodes[e][1:])
                    examdate = bl_examdate + relativedelta(months=+months)
                    exam = examdate.strftime("%m/%d/%Y")

        if exam == "-4":
            print(f"No EXAMDATE for subject %{rid}, at session {viscodes[e]}")

        res_examdates.append(exam)

    return res_examdates


def get_ages(pt_dob, examdates):
    """Calculate age as time passed by since DOB to EXAMDATE.

    :param pt_dob: string - Date of birth of patient ("/%Y" format)
    :param examdates: list - Exam dates ("%m/%d/%Y" format)
    :return: list - Age at each exam date
    """
    from datetime import datetime

    age = []
    dob = datetime.strptime(pt_dob, "/%Y")

    for exam in examdates:
        examdate = datetime.strptime(exam, "%m/%d/%Y")
        delta = examdate - dob
        age.append(round(delta.days / 365.25, 1))

    return age



# coding: utf8

"""This module handles Clinica exceptions."""


class ClinicaException(Exception):
    """Base class for Clinica exceptions."""


class ClinicaMissingDependencyError(ClinicaException):
    """Base class for Clinica dependencies errors."""


class ClinicaBIDSError(ClinicaException):
    """Base class for BIDS errors."""


class ClinicaCAPSError(ClinicaException):
    """Base class for CAPS errors."""


class ClinicaParserError(ClinicaException):
    """Base class for parser errors."""



# coding: utf8


import nibabel as nib
import numpy as np
import pandas as pd


def load_data(image_list, subjects):
    """

    Args:
        image_list:
        subjects:

    Returns:

    """

    subj_average = []
    all_vector = np.array([])
    read_file = pd.io.parsers.read_csv(image_list[0], sep="\t", usecols=[2], header=0)
    read_file = read_file.mean_scalar
    data = np.zeros((len(subjects), len(read_file)))

    for i in range(len(image_list)):
        tsv_file = pd.io.parsers.read_csv(
            image_list[i], sep="\t", usecols=[2], header=0
        )
        subj_average = tsv_file.mean_scalar
        all_vector = np.append(all_vector, subj_average)
    data_temp = np.split(all_vector, len(image_list))

    for i in range(len(image_list)):
        for j in range(len(subj_average)):
            data[i][j] = data_temp[i][j]
    return data


def features_weights(image_list, dual_coefficients, sv_indices, scaler=None):
    """

    Args:
        image_list:
        dual_coefficients:
        sv_indices:
        scaler:

    Returns:

    """

    if len(sv_indices) != len(dual_coefficients):
        raise ValueError(
            f"The number of support vectors indices and the number of coefficients must be the same.\n"
            f"- Number of dual coefficients: {len(dual_coefficients)}\n"
            f"- Number of indices:: {len(sv_indices)}\n"
        )

    if len(image_list) == 0:
        raise ValueError("The number of images must be greater than 0.")

    sv_images = [image_list[i] for i in sv_indices]

    shape = pd.io.parsers.read_csv(sv_images[0], sep="\t", usecols=[2], header=0)
    weights = np.zeros(len(shape))

    for i in range(len(sv_images)):
        subj = pd.io.parsers.read_csv(sv_images[i], sep="\t", usecols=[2], header=0)
        subj_data = subj.mean_scalar
        weights += dual_coefficients[i] * subj_data

    return weights


def weights_to_nifti(weights, atlas, output_filename):
    """

    Args:
        atlas:
        weights:
        output_filename:

    Returns:

    """
    from clinica.utils.atlas import AtlasAbstract

    atlas_path = None
    atlas_classes = AtlasAbstract.__subclasses__()
    for atlas_class in atlas_classes:
        if atlas_class.get_name_atlas() == atlas:
            atlas_path = atlas_class.get_atlas_labels()

    if not atlas_path:
        raise ValueError("Atlas path not found for atlas name " + atlas)

    atlas_image = nib.load(atlas_path)
    atlas_data = atlas_image.get_fdata(dtype="float32")
    labels = list(set(atlas_data.ravel()))
    output_image_weights = np.array(atlas_data, dtype="f")
    for i, n in enumerate(labels):
        index = np.array(np.where(atlas_data == n))
        output_image_weights[index[0, :], index[1, :], index[2, :]] = weights[i]

    affine = atlas_image.get_affine()
    header = atlas_image.get_header()
    output_image = nib.Nifti1Image(output_image_weights, affine, header)
    nib.save(output_image, output_filename)



# coding: utf8


# Functions used by nipype interface.

# Initiate the pipeline
def init_input_node(pet):
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    # Extract image ID
    image_id = get_subject_id(pet)
    print_begin_image(image_id)
    return pet


# Concatenate two transformation in one transformation list
def concatenate_transforms(pet_to_t1w_tranform, t1w_to_mni_tranform):
    """Concatenate two input transformation files into a list.
    Args:
       transform1 (str): first transformation to apply
       transform2 (str): second transformation to apply
    Returns:
       transform_list (list of string): both transform files path in a list
    """
    return [t1w_to_mni_tranform, pet_to_t1w_tranform]


# Normalize the images based on the reference mask region
def suvr_normalization(input_img, norm_img, ref_mask):
    """Normalize the input image according to the reference region.
    It uses nilearn `resample_to_img` and scipy `trim_mean` functions.
    This function is different than the one in other PET pipelines
    because there is a downsampling step.
    Args:
       input_img (str): image to be processed
       norm_img (str): image used to compute the mean of the reference region
       ref_mask (str): mask of the reference region
    Returns:
       output_img (nifty image): normalized nifty image
       mask_template (nifty image): output mask on disk
    """

    import os

    import nibabel as nib
    import numpy as np
    from nilearn.image import resample_to_img
    from scipy.stats import trim_mean

    pet = nib.load(input_img)
    norm = nib.load(norm_img)
    mask = nib.load(ref_mask)

    # Downsample the pet image used for normalization so we can multiply it with the mask
    ds_img = resample_to_img(norm, mask, interpolation="nearest")

    # Compute the mean of the region
    region = np.multiply(ds_img.get_fdata(), mask.get_fdata(dtype="float32"))
    array_region = np.where(region != 0, region, np.nan).flatten()
    region_mean = trim_mean(array_region[~np.isnan(array_region)], 0.1)

    from clinica.utils.stream import cprint

    cprint(region_mean)

    # Divide the value of the image voxels by the computed mean
    data = pet.get_fdata(dtype="float32") / region_mean

    # Create and save the normalized image
    output_img = os.path.join(
        os.getcwd(),
        os.path.basename(input_img).split(".nii")[0] + "_suvr_normalized.nii.gz",
    )

    normalized_img = nib.Nifti1Image(data, pet.affine, header=pet.header)
    normalized_img.to_filename(output_img)

    return output_img


# It crops an image based on the reference.
def crop_nifti(input_img, ref_crop):
    """Crop input image based on the reference. It uses nilearn
    `resample_to_img` function.
    Args:
       input_img (str): image to be processed
       ref_img (str): template used to crop the image
    Returns:
       output_img (nifty image): crop image on disk.
       crop_template (nifty image): output template on disk.
    """

    import os

    import nibabel as nib
    import numpy as np
    from nilearn.image import resample_to_img

    basedir = os.getcwd()

    # resample the individual MRI into the cropped template image
    crop_img = resample_to_img(input_img, ref_crop, force_resample=True)

    output_img = os.path.join(
        basedir, os.path.basename(input_img).split(".nii")[0] + "_cropped.nii.gz"
    )

    crop_img.to_filename(output_img)

    return output_img


def rename_into_caps(
    in_bids_pet,
    fname_pet,
    fname_trans,
    suvr_reference_region,
    uncropped_image,
    fname_pet_in_t1w=None,
):
    """
    Rename the outputs of the pipelines into CAPS format.
    Args:
        in_bids_pet (str): Input BIDS PET to extract the <source_file>
        fname_pet (str): Preprocessed PET file.
        fname_trans (str): Transformation file from PET to MRI space
        suvr_reference_region (str): SUVR mask name for file name output
        uncropped_image (bool): Pipeline argument for image cropping
        fname_pet_in_t1w (bool): Pipeline argument for saving intermediate file
    Returns:
        The different outputs in CAPS format
    """
    import os

    from nipype.interfaces.utility import Rename
    from nipype.utils.filemanip import split_filename

    _, source_file_pet, _ = split_filename(in_bids_pet)

    # Rename into CAPS PET:
    rename_pet = Rename()
    rename_pet.inputs.in_file = fname_pet
    if not uncropped_image:
        suffix = f"_space-MNI152NLin2009cSym_desc-Crop_res-1x1x1_suvr-{suvr_reference_region}_pet.nii.gz"
        rename_pet.inputs.format_string = source_file_pet + suffix
    else:
        suffix = f"_space-MNI152NLin2009cSym_res-1x1x1_suvr-{suvr_reference_region}_pet.nii.gz"
        rename_pet.inputs.format_string = source_file_pet + suffix
    out_caps_pet = rename_pet.run().outputs.out_file

    # Rename into CAPS transformation file:
    rename_trans = Rename()
    rename_trans.inputs.in_file = fname_trans
    rename_trans.inputs.format_string = source_file_pet + "_space-T1w_rigid.mat"
    out_caps_trans = rename_trans.run().outputs.out_file

    # Rename intermediate PET in T1w MRI space
    if fname_pet_in_t1w is not None:
        rename_pet_in_t1w = Rename()
        rename_pet_in_t1w.inputs.in_file = fname_pet_in_t1w
        rename_pet_in_t1w.inputs.format_string = (
            source_file_pet + "_space-T1w_pet.nii.gz"
        )
        out_caps_pet_in_t1w = rename_pet_in_t1w.run().outputs.out_file
    else:
        out_caps_pet_in_t1w = None

    return out_caps_pet, out_caps_trans, out_caps_pet_in_t1w


def print_end_pipeline(pet, final_file):
    """
    Display end message for <subject_id> when <final_file> is connected.
    """
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(pet))



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os.path as op

from nipype.interfaces.base import File, TraitedSpec, Undefined, traits
from nipype.interfaces.mrtrix3.base import MRTrix3Base, MRTrix3BaseInputSpec


class FitTensorInputSpec(MRTrix3BaseInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="input diffusion weighted images",
    )
    out_file = File(
        "dti.mif",
        argstr="%s",
        mandatory=True,
        position=-1,
        usedefault=True,
        desc="the output diffusion tensor image",
    )

    # General options
    in_mask = File(
        exists=True,
        argstr="-mask %s",
        desc=(
            "only perform computation within the specified " "binary brain mask image"
        ),
    )
    method = traits.Enum(
        "nonlinear",
        "loglinear",
        "sech",
        "rician",
        argstr="-method %s",
        desc=("select method used to perform the fitting"),
    )
    reg_term = traits.Float(
        5.0e3,
        usedefault=True,
        argstr="-regularisation %f",
        desc=(
            "specify the strength of the regularisation term on the "
            "magnitude of the tensor elements (default = 5000). This "
            "only applies to the non-linear methods"
        ),
    )


class FitTensorOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output DTI file")


class FitTensor(MRTrix3Base):
    """
    Convert diffusion-weighted images to tensor images


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> tsr = mrt.FitTensor()
    >>> tsr.inputs.in_file = 'dwi.mif'
    >>> tsr.inputs.in_mask = 'mask.nii.gz'
    >>> tsr.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> tsr.cmdline                               # doctest: +ELLIPSIS
    'dwi2tensor -fslgrad bvecs bvals -mask mask.nii.gz \
-regularisation 5000.000000 dwi.mif dti.mif'
    >>> tsr.run()                                 # doctest: +SKIP
    """

    _cmd = "dwi2tensor"
    input_spec = FitTensorInputSpec
    output_spec = FitTensorOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs


# From this line, everything that is defined was to avoid waiting for nipype's
# team to merge @jguillon's requests in the `master` branch and also to avoid
# having to update clinica with the last version of nipype.


class EstimateFODInputSpec(MRTrix3BaseInputSpec):
    algorithm = traits.Enum(
        "csd",
        "msmt_csd",
        argstr="%s",
        position=-8,
        mandatory=True,
        desc="FOD algorithm",
    )
    in_file = File(
        exists=True, argstr="%s", position=-7, mandatory=True, desc="input DWI image"
    )
    wm_txt = File(
        argstr="%s", position=-6, mandatory=True, desc="WM response text file"
    )
    wm_odf = File(
        "wm.mif",
        argstr="%s",
        position=-5,
        usedefault=True,
        mandatory=True,
        desc="output WM ODF",
    )
    #    gm_txt = File(argstr='%s', position=-4, desc='GM response text file')
    #    gm_odf = File('gm.mif', usedefault=True, argstr='%s',
    #                  position=-3, desc='output GM ODF')
    #    csf_txt = File(argstr='%s', position=-2, desc='CSF response text file')
    #    csf_odf = File('csf.mif', usedefault=True, argstr='%s',
    #                   position=-1, desc='output CSF ODF')
    mask_file = File(exists=True, argstr="-mask %s", desc="mask image")

    # DW Shell selection options
    shell = traits.List(
        traits.Float,
        sep=",",
        argstr="-shell %s",
        desc="specify one or more dw gradient shells",
    )
    max_sh = traits.Int(
        8,
        usedefault=True,
        argstr="-lmax %d",
        desc="maximum harmonic degree of response function",
    )
    in_dirs = File(
        exists=True,
        argstr="-directions %s",
        desc=(
            "specify the directions over which to apply the non-negativity "
            "constraint (by default, the built-in 300 direction set is "
            "used). These should be supplied as a text file containing the "
            "[ az el ] pairs for the directions."
        ),
    )


class EstimateFODOutputSpec(TraitedSpec):
    wm_odf = File(argstr="%s", desc="output WM ODF")
    gm_odf = File(argstr="%s", desc="output GM ODF")
    csf_odf = File(argstr="%s", desc="output CSF ODF")


class EstimateFOD(MRTrix3Base):
    """
    Estimate fibre orientation distributions from diffusion data using spherical deconvolution

    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> fod = mrt.EstimateFOD()
    >>> fod.inputs.algorithm = 'csd'
    >>> fod.inputs.in_file = 'dwi.mif'
    >>> fod.inputs.wm_txt = 'wm.txt'
    >>> fod.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> fod.cmdline                               # doctest: +ELLIPSIS
    'dwi2fod -fslgrad bvecs bvals -lmax 8 csd dwi.mif wm.txt wm.mif gm.mif csf.mif'
    >>> fod.run()                                 # doctest: +SKIP
    """

    _cmd = "dwi2fod"
    input_spec = EstimateFODInputSpec
    output_spec = EstimateFODOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["wm_odf"] = op.abspath(self.inputs.wm_odf)
        #        if self.inputs.gm_odf != Undefined:
        #            outputs['gm_odf'] = op.abspath(self.inputs.gm_odf)
        #        if self.inputs.csf_odf != Undefined:
        #            outputs['csf_odf'] = op.abspath(self.inputs.csf_odf)
        return outputs



# coding: utf8


def load_data(mgh_list):
    """

    Args: mgh_list : list of mgh files. Each element contains as many paths as
    needed (each element must be associated to a single subject). Surfaces must
    have the same number of vertices across subjects.

    Returns: data : matrix of raw data

    """
    import nibabel as nib
    import numpy as np

    # Construct 0-matrix with the good size, based on the size of the surfaces
    # provided by the first subject
    N_vertex = (
        []
    )  # array containing the surface size of the different surfaces of a subject
    sample = mgh_list[0]
    for i in range(len(sample)):
        N_vertex.append(np.max(nib.load(sample[i]).header.get_data_shape()))
    data = np.zeros((len(mgh_list), np.sum(N_vertex)))

    # Fill data matrix
    N_cumul = np.concatenate(([0], np.cumsum(N_vertex)))
    for s in range(len(mgh_list)):
        for h in range(len(mgh_list[s])):
            current_data = np.squeeze(
                nib.load(mgh_list[s][h]).get_fdata(dtype="float32")
            )
            data[s, N_cumul[h] : N_cumul[h + 1]] = current_data
    return data



# coding: utf8


def init_input_node(t1w, recon_all_args, output_dir):
    """Initialize the pipeline.

    This function will:
        - Extract <image_id> (e.g. sub-CLNC01_ses-M00) T1w filename;
        - Check FOV of T1w;
        - Create SUBJECTS_DIR for recon-all (otherwise, the command won't run);
        - Print begin execution message.
    """
    import os

    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.freesurfer import check_flags
    from clinica.utils.ux import print_begin_image

    # Extract <image_id>
    image_id = get_subject_id(t1w)

    # Check flags for T1w
    flags = check_flags(t1w, recon_all_args)

    # Create SUBJECTS_DIR for recon-all (otherwise, the command won't run)
    subjects_dir = os.path.join(output_dir, image_id)
    os.makedirs(subjects_dir, exist_ok=True)

    print_begin_image(image_id, ["ReconAllArgs"], [flags])

    return image_id, t1w, flags, subjects_dir


def write_tsv_files(subjects_dir, image_id):
    """
    Generate statistics TSV files in `subjects_dir`/regional_measures folder for `image_id`.

    Notes:
        We do not need to check the line "finished without error" in scripts/recon-all.log.
        If an error occurs, it will be detected by Nipype and the next nodes (including
        write_tsv_files will not be called).
    """
    import os

    from clinica.utils.freesurfer import generate_regional_measures
    from clinica.utils.stream import cprint

    if os.path.isfile(os.path.join(subjects_dir, image_id, "mri", "aparc+aseg.mgz")):
        generate_regional_measures(subjects_dir, image_id)
    else:
        cprint(
            msg=(
                f"{image_id.replace('_', ' | ')} does not contain "
                f"mri/aseg+aparc.mgz file. Creation of regional_measures/ folder will be skipped."
            ),
            lvl="warning",
        )
    return image_id


def save_to_caps(source_dir, image_id, caps_dir, overwrite_caps=False):
    """Save `source_dir`/`image_id`/ to CAPS folder.

    This function copies FreeSurfer segmentation and regional_measures folder of `source_dir`/`image_id`/ to
    `caps_dir`/subjects/<participant_id>/<session_id>/t1_freesurfer_cross_sectional/
    where `image_id` = <participant_id>_<session_id>.

    The `source_dir`/`image_id`/ folder should contain the following elements:
        - fsaverage, lh.EC_average and rh.EC_average symbolic links automatically generated by recon-all
        - `image_id`/ folder containing the FreeSurfer segmentation
        - regional_measures/ folder containing TSV files

    Notes:
        We do not need to check the line "finished without error" in scripts/recon-all.log.
        If an error occurs, it will be detected by Nipype and the next nodes (i.e.  save_to_caps will not be called).
    """
    import os
    import shutil

    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_end_image

    participant_id = image_id.split("_")[0]
    session_id = image_id.split("_")[1]

    destination_dir = os.path.join(
        os.path.expanduser(caps_dir),
        "subjects",
        participant_id,
        session_id,
        "t1",
        "freesurfer_cross_sectional",
    )

    # Save FreeSurfer segmentation
    representative_file = os.path.join(image_id, "mri", "aparc+aseg.mgz")
    representative_source_file = os.path.join(
        os.path.expanduser(source_dir), image_id, representative_file
    )
    representative_destination_file = os.path.join(destination_dir, representative_file)
    if os.path.isfile(representative_source_file):
        if os.path.isfile(representative_destination_file):
            if overwrite_caps:
                shutil.rmtree(destination_dir)
                shutil.copytree(
                    src=os.path.join(source_dir, image_id, image_id),
                    dst=os.path.join(destination_dir, image_id),
                    symlinks=True,
                )
                shutil.copytree(
                    src=os.path.join(source_dir, image_id, "regional_measures"),
                    dst=os.path.join(destination_dir, "regional_measures"),
                    symlinks=True,
                )
        else:
            shutil.copytree(
                src=os.path.join(source_dir, image_id, image_id),
                dst=os.path.join(destination_dir, image_id),
                symlinks=True,
            )
            shutil.copytree(
                src=os.path.join(source_dir, image_id, "regional_measures"),
                dst=os.path.join(destination_dir, "regional_measures"),
                symlinks=True,
            )
        print_end_image(image_id)
    else:
        cprint(
            msg=f"{image_id.replace('_', ' | ')} does not contain mri/aseg+aparc.mgz file. Copy will be skipped.",
            lvl="warning",
        )
    return image_id



import click

from clinica.iotools.converters import cli_param


@click.command(name="oasis-to-bids")
@cli_param.dataset_directory
@cli_param.clinical_data_directory
@cli_param.bids_directory
def cli(
    dataset_directory: str,
    clinical_data_directory: str,
    bids_directory: str,
) -> None:
    """OASIS to BIDS converter.

    Convert the imaging and clinical data of OASIS (http://oasis-brains.org/), located in DATASET_DIRECTORY and
    CLINICAL_DATA_DIRECTORY respectively, to a BIDS dataset in the target BIDS_DIRECTORY.
    """
    from clinica.iotools.converters.oasis_to_bids.oasis_to_bids import OasisToBids

    oasis_to_bids = OasisToBids()
    oasis_to_bids.convert_images(dataset_directory, bids_directory)
    oasis_to_bids.convert_clinical_data(clinical_data_directory, bids_directory)


if __name__ == "__main__":
    cli()



from os import PathLike
from typing import BinaryIO, Iterable, List, Tuple, Union

from pandas import DataFrame


def find_clinical_data(
    clinical_data_directory: PathLike,
) -> Tuple[DataFrame, DataFrame, DataFrame, DataFrame, DataFrame]:
    from pathlib import Path

    import pandas as pd

    from clinica.utils.stream import cprint

    # Read the xls
    try:
        image_data_file = list(Path(clinical_data_directory).glob("*.csv"))
    except StopIteration:
        raise FileNotFoundError("Imaging collection file not found.")

    # Assign the dfs
    df_list = []
    for i in range(0, len(image_data_file)):
        df_list.append(pd.read_csv(str(image_data_file[i])))
        df_list[i] = df_list[i].set_index(df_list[i].axes[1][0])
        if df_list[i].index.name == "XNAT_PETSESSIONDATA ID":
            df_pet = df_list[i]
        if df_list[i].index.name == "MR ID":
            df_mri = df_list[i]
        if df_list[i].index.name == "Subject":
            df_subject = df_list[i]
        if df_list[i].index.name == "PUP_PUPTIMECOURSEDATA ID":
            df_pup = df_list[i]
        if df_list[i].index.name == "ADRC_ADRCCLINICALDATA ID":
            df_adrc = df_list[i]

    # Warn if a dataframe is missing
    if (
        "df_pup" in locals()
        and "df_mri" in locals()
        and "df_adrc" in locals()
        and "df_pet" in locals()
    ):
        cprint(msg="All clinical data have been found", lvl="info")
    else:
        raise FileNotFoundError("Clinical data not found or incomplete")

    return df_pet, df_mri, df_subject, df_pup, df_adrc


def read_clinical_data(
    clinical_data_directory: PathLike,
) -> Tuple[DataFrame, DataFrame, DataFrame, DataFrame, DataFrame]:
    df_pet, df_mri, df_subject, df_pup, df_adrc = find_clinical_data(
        clinical_data_directory
    )
    dictionnary = {
        "pet": df_pet,
        "mri": df_mri,
        "subject": df_subject,
        "pup": df_pup,
        "adrc": df_adrc,
    }
    return dictionnary


def read_imaging_data(imaging_data_directory: PathLike) -> DataFrame:
    from pathlib import Path

    import pandas as pd

    source_path_series = pd.Series(
        find_imaging_data(imaging_data_directory), name="source_path"
    )

    source_dir_series = source_path_series.apply(lambda x: Path(str(x)).parent).rename(
        "source_dir"
    )

    file_spec_series = source_path_series.apply(lambda x: Path(str(x)).parts[0]).rename(
        "path"
    )

    df_source = pd.concat(
        [
            source_path_series,
            file_spec_series,
            source_dir_series,
            file_spec_series.str.split("_", expand=True),
        ],
        axis=1,
    )
    df_source = df_source.rename(
        {0: "Subject", 1: "modality", 2: "Date"}, axis="columns"
    ).drop_duplicates()

    df_source = df_source.assign(participant_id=lambda df: "sub-" + df.Subject)
    return df_source


def find_imaging_data(path_to_source_data: PathLike) -> Iterable[PathLike]:
    from pathlib import Path

    for path in Path(path_to_source_data).rglob("*.nii.gz"):
        yield str(path.relative_to(path_to_source_data))


def intersect_data(df_source: DataFrame, dict_df: dict) -> Tuple[DataFrame, DataFrame]:
    import pandas as pd

    df_adrc = dict_df["adrc"].assign(
        session_id=lambda df: df.index.map(lambda x: x.split("_")[2])
    )
    df_adrc_small = df_adrc[df_adrc.session_id == "d0000"]
    df_adrc_small = df_adrc_small.merge(
        df_source["Subject"], how="inner", on="Subject"
    ).drop_duplicates()
    df_source = df_source.merge(
        df_adrc_small[["Subject", "ageAtEntry"]], how="inner", on="Subject"
    )
    df_source = df_source.assign(
        age=lambda df: (df["ageAtEntry"]) + df["Date"].str[1:].astype("float") / 365.25
    )

    df_subject_small = (
        dict_df["subject"]
        .merge(df_source["Subject"], how="inner", on="Subject")
        .drop_duplicates()
    )
    df_source = df_source.merge(
        dict_df["mri"]["Scanner"], how="left", left_on="path", right_on="MR ID"
    )

    df_small = df_subject_small.merge(df_adrc_small, how="inner", on="Subject")
    df_small = df_small.assign(participant_id=lambda df: "sub-" + df.Subject)
    df_source = df_source.assign(
        session=lambda df: (round(df["Date"].str[1:].astype("int") / (365.25 / 2)) * 6)
    )

    df_source = df_source.assign(
        ses=lambda df: df.session.apply(
            lambda x: f"ses-M{ (5-(len(str(x)))) * '0' + str(int(x))}"
        )
    )

    df_source = df_source.join(
        df_source.modality.map(
            {
                "MR": {"datatype": "anat", "suffix": "T1w"},
                "FDG": {"datatype": "pet", "suffix": "pet", "trc_label": "18FFDG"},
                "PIB": {"datatype": "pet", "suffix": "pet", "trc_label": "11CPIB"},
                "AV45": {"datatype": "pet", "suffix": "pet", "trc_label": "18FAV45"},
            }
        ).apply(pd.Series)
    )
    if "trc_label" in df_source.columns:
        df_source = df_source.assign(
            filename=lambda df: df.apply(
                lambda x: f"{x.participant_id}/{x.ses}/{x.datatype}/"
                f"{x.participant_id}_{x.ses}"
                f"{'_trc-'+x.trc_label if pd.notna(x.trc_label) else ''}"
                f"_{x.suffix}.nii.gz",
                axis=1,
            )
        )
    else:
        df_source = df_source.assign(
            filename=lambda df: df.apply(
                lambda x: f"{x.participant_id}/{x.ses}/{x.datatype}/"
                f"{x.participant_id}_{x.ses}"
                f"_{x.suffix}.nii.gz",
                axis=1,
            )
        )
    df_adrc = df_adrc.merge(df_source["Subject"], how="inner", on="Subject")
    df_adrc = df_adrc.assign(
        session=lambda df: round(df["session_id"].str[1:].astype("int") / (364.25 / 2))
        * 6
    )
    df_adrc = df_adrc.drop_duplicates().set_index(["Subject", "session_id"])
    df_source = df_source.merge(
        df_adrc[
            [
                "session",
                "mmse",
                "cdr",
                "commun",
                "dx1",
                "homehobb",
                "judgment",
                "memory",
                "orient",
                "perscare",
                "sumbox",
                "apoe",
            ]
        ],
        how="left",
        on="session",
    )
    return df_source, df_small


def dataset_to_bids(
    df_source: DataFrame, df_small: DataFrame
) -> Tuple[DataFrame, DataFrame, DataFrame]:
    # Build participants dataframe
    df_participants = (
        df_small[["participant_id", "ageAtEntry", "M/F", "Hand", "Education"]]
        .rename(
            columns={
                "ageAtEntry": "age",
                "M/F": "sex",
                "Hand": "handedness",
            }
        )
        .set_index("participant_id", verify_integrity=True)
    )

    # Build sessions dataframe
    df_session = (
        df_source[
            [
                "participant_id",
                "ses",
                "Date",
                "age",
                "mmse",
                "cdr",
                "commun",
                "dx1",
                "homehobb",
                "judgment",
                "memory",
                "orient",
                "perscare",
                "sumbox",
                "apoe",
            ]
        ]
        .rename(
            columns={
                "ses": "session_id",
                "Date": "source_session_id",
            }
        )
        .astype({"age": "int"})
        .drop_duplicates(subset=["participant_id", "session_id"])
        .set_index(["participant_id", "session_id"], verify_integrity=True)
    )

    # Build scans dataframe
    df_scan = (
        df_source[["filename", "source_dir"]]
        .drop_duplicates(subset=["filename"])
        .set_index("filename", verify_integrity=True)
    )

    return df_participants, df_session, df_scan


def write_to_tsv(dataframe: DataFrame, buffer: Union[PathLike, BinaryIO]) -> None:
    # Save dataframe as a BIDS-compliant TSV file.
    dataframe.to_csv(buffer, sep="\t", na_rep="n/a", date_format="%Y-%m-%d")


def install_bids(sourcedata_dir: PathLike, bids_filename: PathLike) -> None:
    from pathlib import Path

    from fsspec.implementations.local import LocalFileSystem

    fs = LocalFileSystem(auto_mkdir=True)

    source_file = fs.open(fs.ls(sourcedata_dir)[0], mode="rb")
    target_file = fs.open(bids_filename, mode="wb")

    with source_file as sf, target_file as tf:
        tf.write(sf.read())

    source_basename = Path(Path(Path(fs.ls(sourcedata_dir)[0]).stem).stem)
    target_basename = Path(bids_filename.stem).stem

    # The following part adds the sidecar files related to the nifti with the same name: it can be tsv or json files.
    # It may or may not be used, since there might not be any sidecars.
    sidecar_dir = sourcedata_dir.parent / "BIDS"
    for source_sidecar in sidecar_dir.rglob(f"{source_basename}*"):
        target_sidecar = Path.joinpath(bids_filename.parent, target_basename).with_name(
            f"{target_basename}{source_sidecar.suffix}"
        )
        source_file = fs.open(source_sidecar, mode="rb")
        target_file = fs.open(target_sidecar, mode="wb")
        with source_file as sf, target_file as tf:
            tf.write(sf.read())


def write_bids(
    to: PathLike,
    participants: DataFrame,
    sessions: DataFrame,
    scans: DataFrame,
    dataset_directory: PathLike,
) -> List[PathLike]:
    from pathlib import Path

    from fsspec.implementations.local import LocalFileSystem

    to = Path(to)
    fs = LocalFileSystem(auto_mkdir=True)

    # Ensure BIDS hierarchy is written first.
    with fs.transaction:
        with fs.open(to / "participants.tsv", "wb") as participant_file:
            write_to_tsv(participants, participant_file)

        for participant_id, sessions_group in sessions.groupby("participant_id"):
            sessions_group = sessions_group.droplevel("participant_id")
            sessions_filepath = to / participant_id / f"{participant_id}_sessions.tsv"
            with fs.open(sessions_filepath, "wb") as sessions_file:
                write_to_tsv(sessions_group, sessions_file)

    # Perform import of imaging data next.
    for filename, metadata in scans.iterrows():
        path = Path(dataset_directory) / metadata.source_dir
        install_bids(sourcedata_dir=path, bids_filename=to / filename)
    return scans.index.to_list()



from typing import Optional

import click

from clinica.pipelines import cli_param

pipeline_name = "machinelearning-prepare-spatial-svm"


@click.command(pipeline_name)
@cli_param.argument.caps_directory
@cli_param.argument.group_label
@cli_param.argument.orig_input_data_ml
@cli_param.option_group.custom_pipeline_options(
    "Pipeline options if you use inputs from pet-volume pipeline"
)
@cli_param.option.acq_label
@cli_param.option.suvr_reference_region
@cli_param.option.use_pvc_data
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.n_procs
@cli_param.option_group.advanced_pipeline_options
@cli_param.option_group.option(
    "-fwhm",
    "--full_width_half_maximum",
    default=4.0,
    help=(
        "Amount of regularization (in mm). In practice, we found the default value "
        "(--full_width_half_maximum %(default)s) to be optimal. We therefore "
        "do not recommend to change it unless you have a specific reason to do so."
    ),
)
def cli(
    caps_directory: str,
    group_label: str,
    orig_input_data_ml: str,
    acq_label: Optional[str] = None,
    suvr_reference_region: Optional[str] = None,
    use_pvc_data: bool = False,
    full_width_half_maximum: float = 4.0,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Prepare input data for SVM with spatial and anatomical regularization.

       GROUP_LABEL is an user-defined identifier to target a specific group of subjects.

       The third positional argument can be t1-volume to use tissue maps or pet-volume to use standardized uptake value ratio (SUVR) maps.

    Prerequisites: You need to execute the t1-volume pipeline to run the pipeline on T1-weighted MRI data, and the t1-volume + pet-volume pipelines to apply it to PET data.

    See https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/MachineLearning_PrepareSVM/"
    """
    from networkx import Graph

    from clinica.utils.exceptions import ClinicaException
    from clinica.utils.ux import print_end_pipeline

    from .spatial_svm_pipeline import SpatialSVM

    if orig_input_data_ml == "pet-volume":
        if not acq_label:
            raise ClinicaException(
                "You selected pet-volume pipeline without setting --acq_label flag. "
                "Clinica will now exit."
            )
        if not suvr_reference_region:
            raise ClinicaException(
                "You selected pet-volume pipeline without setting --suvr_reference_region flag. "
                "Clinica will now exit."
            )

    parameters = {
        # Clinica compulsory arguments
        "group_label": group_label,
        "orig_input_data_ml": orig_input_data_ml,
        # Optional arguments for inputs from pet-volume pipeline
        "acq_label": acq_label,
        "use_pvc_data": use_pvc_data,
        "suvr_reference_region": suvr_reference_region,
        # Advanced arguments
        "fwhm": full_width_half_maximum,
    }

    pipeline = SpatialSVM(
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



from typing import List, Optional

import click

from clinica.pipelines import cli_param
from clinica.utils.pet import LIST_SUVR_REFERENCE_REGIONS

pipeline_name = "deeplearning-prepare-data"


@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@click.argument(
    "modality",
    type=click.Choice(["t1-linear", "t1-extensive", "pet-linear", "custom"]),
    default="t1-linear",
)
@click.argument(
    "extract-method",
    type=click.Choice(["image", "slice", "patch", "roi"]),
    default="image",
)
@cli_param.option_group.pipeline_specific_options
@cli_param.option_group.option(
    "-uui",
    "--use_uncropped_image",
    is_flag=True,
    help="Use the uncropped image instead of the cropped image generated by t1-linear or pet-linear.",
)
@cli_param.option_group.custom_pipeline_options(
    "Pipeline options if you chose `patch` extraction"
)
@cli_param.option_group.option(
    "-ps",
    "--patch_size",
    default=50,
    show_default=True,
    help="Patch size if EXTRACT_METHOD is `patch`.",
)
@cli_param.option_group.option(
    "-ss",
    "--stride_size",
    default=50,
    show_default=True,
    help="Stride size if EXTRACT_METHOD is ``patch.",
)
@cli_param.option_group.custom_pipeline_options(
    "Pipeline options if you chose `slice` extraction"
)
@cli_param.option_group.option(
    "-sd",
    "--slice_direction",
    type=click.IntRange(0, 2),
    default=0,
    show_default=True,
    help="Slice direction if EXTRACT_METHOD is `slice`. 0: Sagittal plane, 1: Coronal plane, 2: Axial plane.",
)
@cli_param.option_group.option(
    "-sm",
    "--slice_mode",
    type=click.Choice(["rgb", "single"]),
    default="rgb",
    show_default=True,
    help=(
        "Slice mode if EXTRACT_METHOD is `slice`. rgb: Save the slice in three identical channels, "
        "single: Save the slice in a single channel."
    ),
)
@cli_param.option_group.custom_pipeline_options(
    "Pipeline options if you chose `roi` extraction"
)
@cli_param.option_group.option(
    "--roi_list",
    multiple=True,
    help="List of regions to be extracted.",
)
@cli_param.option_group.option(
    "--roi_uncrop_output",
    is_flag=True,
    help="Disable cropping option so the output tensors have the same size than the whole image.",
)
@cli_param.option_group.option(
    "-ct",
    "--custom_template",
    help="Name of the template used when modality is set to `custom`.",
)
@cli_param.option_group.option(
    "-cmp",
    "--custom_mask_pattern",
    default="",
    help=(
        "Select only the masks containing this pattern. The mask with the shortest name is taken. "
        "This argument is taken into account only when the modality is set to `custom`."
    ),
)
@cli_param.option_group.custom_pipeline_options(
    "Pipeline options if you chose `pet-linear` pipeline"
)
@cli_param.option_group.option(
    "--acq_label",
    type=click.Choice(["av45", "fdg"]),
    help=(
        "Name of the label given to the PET acquisition, specifying  the tracer used (acq-<acq_label>). "
        "For instance it can be 'fdg' for fluorodeoxyglucose or 'av45' for florbetapir. This option only applies to "
        "the `pet-linear` pipeline."
    ),
)
@cli_param.option_group.option(
    "-suvr",
    "--suvr_reference_region",
    type=click.Choice(LIST_SUVR_REFERENCE_REGIONS),
    help=(
        "Intensity normalization using the average PET uptake in reference regions resulting in a standardized uptake "
        "value ratio (SUVR) map. It can be cerebellumPons or cerebellumPon2 (used for amyloid tracers) or pons or "
        "pons2 (used for 18F-FDG tracers). This option only applies to `pet-linear` pipeline."
    ),
)
@cli_param.option_group.custom_pipeline_options(
    "Pipeline options if you chose `custom` modality"
)
@cli_param.option_group.option(
    "-cn",
    "--custom_suffix",
    default="",
    help=(
        "Suffix to append to filenames for a custom modality, for instance "
        "`graymatter_space-Ixi549Space_modulated-off_probability.nii.gz`, or "
        "`segm-whitematter_probability.nii.gz`"
    ),
)
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@cli_param.option.n_procs
def cli(
    caps_directory: str,
    modality: str,
    extract_method: str,
    use_uncropped_image: bool = False,
    patch_size: int = 50,
    stride_size: int = 50,
    slice_direction: int = 0,
    slice_mode: str = "rgb",
    roi_list: Optional[List[str]] = None,
    roi_uncrop_output: bool = False,
    custom_template: Optional[str] = None,
    custom_mask_pattern: str = "",
    acq_label: Optional[str] = None,
    suvr_reference_region: Optional[str] = None,
    custom_suffix: str = "",
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Prepare data generated by Clinica for PyTorch with tensor extraction.

       It currently supports three preprocessed modalities (t1-linear, t1-extensive and pet-linear).

       Extraction of custom modalities is possible using the right argument and the option `--custom_suffix`.

       Extracted tensors can be obtained in three forms: the whole 3D image, 3D patches or 2D slices.

    See https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/DeepLearning_PrepareData/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .deeplearning_prepare_data_pipeline import DeepLearningPrepareData

    parameters = {
        "modality": modality,
        "extract_method": extract_method,
        "patch_size": patch_size,
        "stride_size": stride_size,
        "slice_direction": slice_direction,
        "slice_mode": slice_mode,
        "roi_list": roi_list,
        "roi_uncrop_output": roi_uncrop_output,
        "custom_template": custom_template,
        "custom_mask_pattern": custom_mask_pattern,
        "use_uncropped_image": use_uncropped_image,
        "custom_suffix": custom_suffix,
        "acq_label": acq_label,
        "suvr_reference_region": suvr_reference_region,
    }

    pipeline = DeepLearningPrepareData(
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os.path as op

from nipype.interfaces.base import File, TraitedSpec, traits
from nipype.interfaces.mrtrix3.base import MRTrix3Base, MRTrix3BaseInputSpec


class TractographyInputSpec(MRTrix3BaseInputSpec):
    sph_trait = traits.Tuple(
        traits.Float, traits.Float, traits.Float, traits.Float, argstr="%f,%f,%f,%f"
    )

    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="input file to be processed",
    )

    out_file = File(
        "tracked.tck",
        argstr="%s",
        mandatory=True,
        position=-1,
        usedefault=True,
        desc="output file containing tracks",
    )

    algorithm = traits.Enum(
        "iFOD2",
        "FACT",
        "iFOD1",
        "Nulldist",
        "SD_Stream",
        "Tensor_Det",
        "Tensor_Prob",
        usedefault=True,
        argstr="-algorithm %s",
        desc="tractography algorithm to be used",
    )

    # ROIs processing options
    roi_incl = traits.Either(
        File(exists=True),
        sph_trait,
        argstr="-include %s",
        desc=(
            "specify an inclusion region of interest, streamlines must"
            " traverse ALL inclusion regions to be accepted"
        ),
    )
    roi_excl = traits.Either(
        File(exists=True),
        sph_trait,
        argstr="-exclude %s",
        desc=(
            "specify an exclusion region of interest, streamlines that"
            " enter ANY exclude region will be discarded"
        ),
    )
    roi_mask = traits.Either(
        File(exists=True),
        sph_trait,
        argstr="-mask %s",
        desc=(
            "specify a masking region of interest. If defined,"
            "streamlines exiting the mask will be truncated"
        ),
    )

    # Streamlines tractography options
    step_size = traits.Float(
        argstr="-step %f",
        desc=(
            "set the step size of the algorithm in mm (default is 0.1"
            " x voxelsize; for iFOD2: 0.5 x voxelsize)"
        ),
    )
    angle = traits.Float(
        argstr="-angle %f",
        desc=(
            "set the maximum angle between successive steps (default "
            "is 90deg x stepsize / voxelsize)"
        ),
    )
    select = traits.Int(
        argstr="-select %d",
        desc=(
            "set the desired number of tracks. The program will continue"
            " to generate tracks until this number of tracks have been "
            "selected and written to the output file"
        ),
    )
    max_tracks = traits.Int(
        argstr="-maxnum %d",
        desc=(
            "set the maximum number of tracks to generate. The program "
            "will not generate more tracks than this number, even if "
            "the desired number of tracks hasn't yet been reached "
            "(default is 100 x number)"
        ),
    )
    max_length = traits.Float(
        argstr="-maxlength %f",
        desc=(
            "set the maximum length of any track in mm (default is " "100 x voxelsize)"
        ),
    )
    min_length = traits.Float(
        argstr="-minlength %f",
        desc=(
            "set the minimum length of any track in mm (default is " "5 x voxelsize)"
        ),
    )
    cutoff = traits.Float(
        argstr="-cutoff %f",
        desc=(
            "set the FA or FOD amplitude cutoff for terminating "
            "tracks (default is 0.1)"
        ),
    )
    cutoff_init = traits.Float(
        argstr="-initcutoff %f",
        desc=(
            "set the minimum FA or FOD amplitude for initiating "
            "tracks (default is the same as the normal cutoff)"
        ),
    )
    n_trials = traits.Int(
        argstr="-trials %d",
        desc=(
            "set the maximum number of sampling trials at each point"
            " (only used for probabilistic tracking)"
        ),
    )
    unidirectional = traits.Bool(
        argstr="-unidirectional",
        desc=(
            "track from the seed point in one direction only "
            "(default is to track in both directions)"
        ),
    )
    init_dir = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        argstr="-initdirection %f,%f,%f",
        desc=(
            "specify an initial direction for the tracking (this "
            "should be supplied as a vector of 3 comma-separated values"
        ),
    )
    noprecompt = traits.Bool(
        argstr="-noprecomputed",
        desc=(
            "do NOT pre-compute legendre polynomial values. Warning: this "
            "will slow down the algorithm by a factor of approximately 4"
        ),
    )
    power = traits.Int(
        argstr="-power %d",
        desc=("raise the FOD to the power specified (default is 1/nsamples)"),
    )
    n_samples = traits.Int(
        4,
        usedefault=True,
        argstr="-samples %d",
        desc=(
            "set the number of FOD samples to take per step for the 2nd "
            "order (iFOD2) method"
        ),
    )
    use_rk4 = traits.Bool(
        argstr="-rk4",
        desc=(
            "use 4th-order Runge-Kutta integration (slower, but eliminates"
            " curvature overshoot in 1st-order deterministic methods)"
        ),
    )
    stop = traits.Bool(
        argstr="-stop",
        desc=(
            "stop propagating a streamline once it has traversed all " "include regions"
        ),
    )
    downsample = traits.Float(
        argstr="-downsample %f",
        desc="downsample the generated streamlines to reduce output file size",
    )

    # Anatomically-Constrained Tractography options
    act_file = File(
        exists=True,
        argstr="-act %s",
        desc=(
            "use the Anatomically-Constrained Tractography framework during"
            " tracking; provided image must be in the 5TT "
            "(five - tissue - type) format"
        ),
    )
    backtrack = traits.Bool(argstr="-backtrack", desc="allow tracks to be truncated")

    crop_at_gmwmi = traits.Bool(
        argstr="-crop_at_gmwmi",
        desc=(
            "crop streamline endpoints more "
            "precisely as they cross the GM-WM interface"
        ),
    )

    # Tractography seeding options
    seed_sphere = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        traits.Float,
        argstr="-seed_sphere %f,%f,%f,%f",
        desc="spherical seed",
    )
    seed_image = File(
        exists=True,
        argstr="-seed_image %s",
        desc="seed streamlines entirely at random within mask",
    )
    seed_rnd_voxel = traits.Tuple(
        File(exists=True),
        traits.Int(),
        argstr="-seed_random_per_voxel %s %d",
        xor=["seed_image", "seed_grid_voxel"],
        desc=(
            "seed a fixed number of streamlines per voxel in a mask "
            "image; random placement of seeds in each voxel"
        ),
    )
    seed_grid_voxel = traits.Tuple(
        File(exists=True),
        traits.Int(),
        argstr="-seed_grid_per_voxel %s %d",
        xor=["seed_image", "seed_rnd_voxel"],
        desc=(
            "seed a fixed number of streamlines per voxel in a mask "
            "image; place seeds on a 3D mesh grid (grid_size argument "
            "is per axis; so a grid_size of 3 results in 27 seeds per"
            " voxel)"
        ),
    )
    seed_rejection = File(
        exists=True,
        argstr="-seed_rejection %s",
        desc=(
            "seed from an image using rejection sampling (higher "
            "values = more probable to seed from"
        ),
    )
    seed_gmwmi = File(
        exists=True,
        argstr="-seed_gmwmi %s",
        requires=["act_file"],
        desc=(
            "seed from the grey matter - white matter interface (only "
            "valid if using ACT framework)"
        ),
    )
    seed_dynamic = File(
        exists=True,
        argstr="-seed_dynamic %s",
        desc=(
            "determine seed points dynamically using the SIFT model "
            "(must not provide any other seeding mechanism). Note that"
            " while this seeding mechanism improves the distribution of"
            " reconstructed streamlines density, it should NOT be used "
            "as a substitute for the SIFT method itself."
        ),
    )
    max_seed_attempts = traits.Int(
        argstr="-max_seed_attempts %d",
        desc=(
            "set the maximum number of times that the tracking "
            "algorithm should attempt to find an appropriate tracking"
            " direction from a given seed point"
        ),
    )
    out_seeds = File(
        "out_seeds.nii.gz",
        usedefault=True,
        argstr="-output_seeds %s",
        desc=("output the seed location of all successful streamlines to" " a file"),
    )


class TractographyOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output filtered tracks")
    out_seeds = File(
        desc=("output the seed location of all successful" " streamlines to a file")
    )


class Tractography(MRTrix3Base):
    """
    Performs streamlines tractography after selecting the appropriate
    algorithm.

    .. [FACT] Mori, S.; Crain, B. J.; Chacko, V. P. & van Zijl,
      P. C. M. Three-dimensional tracking of axonal projections in the
      brain by magnetic resonance imaging. Annals of Neurology, 1999,
      45, 265-269

    .. [iFOD1] Tournier, J.-D.; Calamante, F. & Connelly, A. MRtrix:
      Diffusion tractography in crossing fiber regions. Int. J. Imaging
      Syst. Technol., 2012, 22, 53-66

    .. [iFOD2] Tournier, J.-D.; Calamante, F. & Connelly, A. Improved
      probabilistic streamlines tractography by 2nd order integration
      over fibre orientation distributions. Proceedings of the
      International Society for Magnetic Resonance in Medicine, 2010, 1670

    .. [Nulldist] Morris, D. M.; Embleton, K. V. & Parker, G. J.
      Probabilistic fibre tracking: Differentiation of connections from
      chance events. NeuroImage, 2008, 42, 1329-1339

    .. [Tensor_Det] Basser, P. J.; Pajevic, S.; Pierpaoli, C.; Duda, J.
      and Aldroubi, A. In vivo fiber tractography using DT-MRI data.
      Magnetic Resonance in Medicine, 2000, 44, 625-632

    .. [Tensor_Prob] Jones, D. Tractography Gone Wild: Probabilistic Fibre
      Tracking Using the Wild Bootstrap With Diffusion Tensor MRI. IEEE
      Transactions on Medical Imaging, 2008, 27, 1268-1274


    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> tk = mrt.Tractography()
    >>> tk.inputs.in_file = 'fods.mif'
    >>> tk.inputs.roi_mask = 'mask.nii.gz'
    >>> tk.inputs.seed_sphere = (80, 100, 70, 10)
    >>> tk.cmdline                               # doctest: +ELLIPSIS
    'tckgen -algorithm iFOD2 -samples 4 -output_seeds out_seeds.nii.gz \
-mask mask.nii.gz -seed_sphere \
80.000000,100.000000,70.000000,10.000000 fods.mif tracked.tck'
    >>> tk.run()                                 # doctest: +SKIP
    """

    _cmd = "tckgen"
    input_spec = TractographyInputSpec
    output_spec = TractographyOutputSpec

    def _format_arg(self, name, trait_spec, value):
        if "roi_" in name and isinstance(value, tuple):
            value = ["%f" % v for v in value]
            return trait_spec.argstr % ",".join(value)

        return super(Tractography, self)._format_arg(name, trait_spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = op.abspath(self.inputs.out_file)
        return outputs



# coding: utf8
"""This module contains functions used for the registration aspects."""

from typing import Optional


def convert_flirt_transformation_to_mrtrix_transformation(
    in_source_image: str,
    in_reference_image: str,
    in_flirt_matrix: str,
    name_output_matrix: Optional[str] = None,
) -> str:
    """Convert flirt matrix to mrtrix matrix.

    This function converts a transformation matrix produced by FSL's flirt
    command into a format usable by MRtrix. The output of this function
    is usually for the mrtransform command.

    Args:
        in_source_image (str): File containing the source image used in
            FSL flirt with the -in flag.
        in_reference_image (str): File containing the reference image used in
            FSL flirt with the -ref flag.
        in_flirt_matrix (str): File containing the transformation matrix
            obtained by FSL flirt.
        name_output_matrix (str, optional): Name of the output matrix.
            Defaults to "mrtrix_matrix.mat".

    Returns:
        out_mrtrix_matrix (str): Transformation matrix in MRtrix format.
    """
    import os

    from clinica.utils.check_dependency import check_mrtrix

    check_mrtrix()

    assert os.path.isfile(in_source_image)
    assert os.path.isfile(in_reference_image)
    assert os.path.isfile(in_flirt_matrix)

    out_mrtrix_matrix = os.path.abspath(name_output_matrix or "mrtrix_matrix.mat")

    cmd = f"transformconvert {in_flirt_matrix} {in_source_image} {in_reference_image} flirt_import {out_mrtrix_matrix}"
    os.system(cmd)

    return out_mrtrix_matrix


def apply_ants_registration_syn_quick_transformation(
    in_image: str,
    in_reference_image: str,
    in_affine_transformation: str,
    in_bspline_transformation: str,
    name_output_image: Optional[str] = None,
):
    """Apply a transformation obtained with antsRegistrationSyNQuick.sh.

    This function applies a rigid & deformable B-Spline syn transformation
    which has been estimated previously with antsRegistrationSyNQuick script.

    Args:
        in_image (str): File containing the input image to be transformed.
        in_reference_image (str): File defining the spacing, origin, size,
            and direction of the output warped image.
        in_affine_transformation (str): File containing the transformation
            matrix obtained by antsRegistrationSyNQuick (expected file:
            [Prefix]0GenericAffine.mat).
        in_bspline_transformation (str): File containing the transformation
            matrix obtained by antsRegistrationSyNQuick (expected file:
            [Prefix]1Warp.nii.gz).
        name_output_image (str, optional): Name of the output image.
            Defaults to "deformed_image.nii.gz".

    Returns:
        out_deformed_image (str): File containing the deformed image according
            to in_affine_transformation and in_bspline_transformation
            transformations.
    """
    import os

    from clinica.utils.check_dependency import check_ants

    check_ants()

    assert os.path.isfile(in_image)
    assert os.path.isfile(in_affine_transformation)
    assert os.path.isfile(in_bspline_transformation)

    out_deformed_image = os.path.abspath(name_output_image or "deformed_image.nii.gz")

    cmd = (
        f"antsApplyTransforms -d 3 -e 0 -i {in_image} -o {out_deformed_image} "
        f"-t {in_bspline_transformation} -t {in_affine_transformation} "
        f"-r {in_reference_image} --interpolation Linear"
    )
    os.system(cmd)

    return out_deformed_image


def ants_registration_syn_quick(
    fix_image, moving_image, prefix_output: Optional[str] = None
):
    """Small wrapper for antsRegistrationSyNQuick.sh.

    This function calls antsRegistrationSyNQuick.sh in order to register
    non-linearly moving_image towards fix_image.

    Args:
        fix_image (str): The target image.
        moving_image (str): The source image.
        prefix_output (str, optional): Prefix for output files.
            Defaults to "SyN_Quick".
            (format: <prefix_output>[Warped|0GenericAffine|1Warp|
            InverseWarped|1InverseWarp])

    Returns:
        The deformed image with the deformation parameters.
    """
    import os

    from clinica.utils.check_dependency import check_ants

    check_ants()

    prefix_output = prefix_output or "SyN_Quick"
    image_warped = os.path.abspath(prefix_output + "Warped.nii.gz")
    affine_matrix = os.path.abspath(prefix_output + "0GenericAffine.mat")
    warp = os.path.abspath(prefix_output + "1Warp.nii.gz")
    inverse_warped = os.path.abspath(prefix_output + "InverseWarped.nii.gz")
    inverse_warp = os.path.abspath(prefix_output + "1InverseWarp.nii.gz")

    cmd = f"antsRegistrationSyNQuick.sh -t b -d 3 -f {fix_image} -m {moving_image} -o {prefix_output}"
    os.system(cmd)

    return image_warped, affine_matrix, warp, inverse_warped, inverse_warp


def ants_combine_transform(in_file, transforms_list, reference):
    """Apply a transformation obtained with antsRegistrationSyNQuick.sh.

    This function applies a rigid & deformable B-Spline syn transformation
    which has been estimated previously with antsRegistrationSyNQuick script.

    Args:
        in_file (str): File containing the input image to be transformed.
        reference (str): File defining the spacing, origin, size, and
            direction of the output warped image.
        transforms_list (str): File containing the transformations
            obtained by antsRegistrationSyNQuick.

    Returns:
        out_warp (str): File containing the deformed image according to
            transforms_list in_affine_transformation and
            in_bspline_transformation transformations.
    """
    import os

    from clinica.utils.check_dependency import check_ants

    check_ants()

    out_warp = os.path.abspath("out_warp.nii.gz")

    transforms = ""
    for trans in transforms_list:
        transforms += " " + trans
    cmd = f"antsApplyTransforms -o [out_warp.nii.gz,1] -i {in_file} -r {reference} -t {transforms}"
    os.system(cmd)

    return out_warp



"""Convert the NIFD dataset into BIDS."""

from os import PathLike
from typing import List


def convert_images(
    path_to_dataset: PathLike,
    bids_dir: PathLike,
    path_to_clinical: PathLike,
) -> List[PathLike]:
    """Convert the entire dataset in BIDS.

    Scans available files in the path_to_dataset,
    identifies the patients that have images described by the JSON file,
    converts the image with the highest quality for each category.
    """
    from .oasis3_utils import (
        dataset_to_bids,
        intersect_data,
        read_clinical_data,
        read_imaging_data,
        write_bids,
    )

    # read the clinical data files
    dict_df = read_clinical_data(path_to_clinical)

    # makes a df of the imaging data
    imaging_data = read_imaging_data(path_to_dataset)

    # intersect the data
    imaging_data, df_small = intersect_data(imaging_data, dict_df)

    # build the tsv
    participants, sessions, scans = dataset_to_bids(imaging_data, df_small)

    written = write_bids(
        to=bids_dir,
        participants=participants,
        sessions=sessions,
        scans=scans,
        dataset_directory=path_to_dataset,
    )

    return written



# coding: utf8


import clinica.pipelines.engine as cpe


class T1FreeSurferTemplate(cpe.Pipeline):
    """FreeSurfer Longitudinal template class.

    Returns:
        A clinica pipeline object containing the T1FreeSurferTemplate pipeline.
    """

    @staticmethod
    def get_processed_images(caps_directory, subjects, sessions):
        import os
        import re

        from clinica.utils.input_files import T1_FS_T_DESTRIEUX
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.longitudinal import get_long_id
        from clinica.utils.participant import get_unique_subjects

        [list_participant_id, list_list_session_ids] = get_unique_subjects(
            subjects, sessions
        )
        list_long_id = [
            get_long_id(list_session_ids) for list_session_ids in list_list_session_ids
        ]

        image_ids = []
        if os.path.isdir(caps_directory):
            t1_freesurfer_files = clinica_file_reader(
                list_participant_id,
                list_long_id,
                caps_directory,
                T1_FS_T_DESTRIEUX,
                False,
            )
            image_ids = [
                re.search(r"(sub-[a-zA-Z0-9]+)_(long-[a-zA-Z0-9]+)", file).group()
                for file in t1_freesurfer_files
            ]
        return image_ids

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""

    def check_custom_dependencies(self):
        """Check dependencies that cannot be listed in `info.json`."""

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Note:
            The list of inputs of the T1FreeSurferTemplate pipeline is:
                * participant_id (str): Participant ID (e.g. "sub-CLNC01")
                * list_session_ids (List[str]: List of session IDs associated to `participant_id`
                    (e.g. ["ses-M00"] or ["ses-M00", "ses-M18", "ses-M36"])

        Returns:
            A list of (string) input fields name.
        """
        return ["participant_id", "list_session_ids"]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Note:
            The list of inputs of the T1FreeSurferTemplate pipeline is:
                * image_id (str): Image ID (e.g. "sub-CLNC01_long-M00M18")

        Returns:
            A list of (string) output fields name.
        """
        return ["image_id"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.filemanip import extract_subjects_sessions_from_filename
        from clinica.utils.input_files import T1_FS_DESTRIEUX
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.longitudinal import (
            get_long_id,
            get_participants_long_id,
            read_sessions,
        )
        from clinica.utils.participant import (
            get_unique_subjects,
            unique_subjects_sessions_to_subjects_sessions,
        )
        from clinica.utils.stream import cprint

        from .longitudinal_utils import (
            extract_participant_long_ids_from_filename,
            save_part_sess_long_ids_to_tsv,
        )

        # Display image(s) already present in CAPS folder
        # ===============================================
        output_ids = self.get_processed_images(
            self.caps_directory, self.subjects, self.sessions
        )
        (
            processed_participants,
            processed_long_sessions,
        ) = extract_participant_long_ids_from_filename(output_ids)
        if len(processed_participants) > 0:
            cprint(
                msg=(
                    f"Clinica found {len(processed_participants)} participant(s) "
                    "already processed in CAPS directory:"
                ),
                lvl="warning",
            )
            for p_id, l_id in zip(processed_participants, processed_long_sessions):
                cprint(f"{p_id} | {l_id}", lvl="warning")
            if self.overwrite_caps:
                output_folder = "<CAPS>/subjects/<participant_id>/<long_id>/freesurfer_unbiased_template/"
                cprint(f"Output folders in {output_folder} will be recreated.", lvl="warning")
            else:
                cprint("Participant(s) will be ignored by Clinica.", lvl="warning")
                input_ids = [
                    p_id + "_" + s_id
                    for p_id, s_id in zip(self.subjects, self.sessions)
                ]
                processed_sessions_per_participant = [
                    read_sessions(self.caps_directory, p_id, l_id)
                    for (p_id, l_id) in zip(
                        processed_participants, processed_long_sessions
                    )
                ]
                participants, sessions = unique_subjects_sessions_to_subjects_sessions(
                    processed_participants, processed_sessions_per_participant
                )
                processed_ids = [
                    p_id + "_" + s_id for p_id, s_id in zip(participants, sessions)
                ]
                to_process_ids = list(set(input_ids) - set(processed_ids))
                self.subjects, self.sessions = extract_subjects_sessions_from_filename(
                    to_process_ids
                )

        # Check that t1-freesurfer has run on the CAPS directory
        try:
            clinica_file_reader(
                self.subjects, self.sessions, self.caps_directory, T1_FS_DESTRIEUX
            )
        except ClinicaException as e:
            err_msg = (
                "Clinica faced error(s) while trying to read files in your CAPS directory.\n"
                + str(e)
            )
            raise ClinicaCAPSError(err_msg)

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        folder_participants_tsv = os.path.join(self.base_dir, self.name)
        long_ids = get_participants_long_id(self.subjects, self.sessions)
        save_part_sess_long_ids_to_tsv(
            self.subjects, self.sessions, long_ids, folder_participants_tsv
        )

        [list_participant_id, list_list_session_ids] = get_unique_subjects(
            self.subjects, self.sessions
        )
        list_long_id = [
            get_long_id(list_session_ids) for list_session_ids in list_list_session_ids
        ]

        def print_images_to_process(
            unique_part_list, per_part_session_list, list_part_long_id
        ):
            cprint(
                f"The pipeline will be run on the following {len(unique_part_list)} participant(s):"
            )
            for (part_id, list_sess_id, list_id) in zip(
                unique_part_list, per_part_session_list, list_part_long_id
            ):
                sessions_participant = ", ".join(s_id for s_id in list_sess_id)
                cprint(f"\t{part_id} | {sessions_participant} | {list_id}")

        if len(self.subjects):
            # TODO: Generalize long IDs to the message display
            print_images_to_process(
                list_participant_id, list_list_session_ids, list_long_id
            )
            cprint(
                "List available in %s"
                % os.path.join(folder_participants_tsv, "participants.tsv")
            )
            cprint("The pipeline will last approximately 10 hours per participant.")

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("participant_id", list_participant_id),
                ("list_session_ids", list_list_session_ids),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        # fmt: off
        self.connect(
            [
                (read_node, self.input_node, [("participant_id", "participant_id")]),
                (read_node, self.input_node, [("list_session_ids", "list_session_ids")]),
            ]
        )
        # fmt: on

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .t1_freesurfer_template_utils import save_to_caps

        save_to_caps = npe.Node(
            interface=nutil.Function(
                input_names=[
                    "source_dir",
                    "image_id",
                    "list_session_ids",
                    "caps_dir",
                    "overwrite_caps",
                ],
                output_names=["image_id"],
                function=save_to_caps,
            ),
            name="SaveToCaps",
        )
        save_to_caps.inputs.source_dir = os.path.join(
            self.base_dir, self.name, "ReconAll"
        )
        save_to_caps.inputs.caps_dir = self.caps_directory
        save_to_caps.inputs.overwrite_caps = False

        # fmt: off
        self.connect(
            [
                (self.input_node, save_to_caps, [("list_session_ids", "list_session_ids")]),
                (self.output_node, save_to_caps, [("image_id", "image_id")]),
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .t1_freesurfer_template_utils import (
            init_input_node,
            move_subjects_dir_to_source_dir,
            run_recon_all_base,
        )

        # Nodes declaration
        # =================
        # Initialize the pipeline
        init_input = npe.Node(
            interface=nutil.Function(
                input_names=[
                    "caps_dir",
                    "participant_id",
                    "list_session_ids",
                    "output_dir",
                ],
                output_names=["image_id", "subjects_dir", "flags"],
                function=init_input_node,
            ),
            name="0-InitPipeline",
        )
        init_input.inputs.caps_dir = self.caps_directory
        init_input.inputs.output_dir = os.path.join(
            self.base_dir, self.name, "ReconAll"
        )

        # Run recon-all command
        recon_all = npe.Node(
            interface=nutil.Function(
                input_names=["subjects_dir", "subject_id", "flags", "directive"],
                output_names=["subject_id"],
                function=run_recon_all_base,
            ),
            name="1-SegmentationReconAll",
        )
        recon_all.inputs.directive = "-all"

        # Move $SUBJECT_DIR to source_dir (1 time point case)
        move_subjects_dir = npe.Node(
            interface=nutil.Function(
                input_names=["subjects_dir", "source_dir", "subject_id"],
                output_names=["subject_id"],
                function=move_subjects_dir_to_source_dir,
            ),
            name="2-MoveSubjectsDir",
        )
        move_subjects_dir.inputs.source_dir = os.path.join(
            self.base_dir, self.name, "ReconAll"
        )

        # Connections
        # ===========
        # fmt: off
        self.connect(
            [
                # Initialize the pipeline
                (self.input_node, init_input, [("participant_id", "participant_id")]),
                (self.input_node, init_input, [("list_session_ids", "list_session_ids")],
                ),
                # Run recon-all command
                (init_input, recon_all, [("subjects_dir", "subjects_dir"),
                                         ("image_id", "subject_id"),
                                         ("flags", "flags")]),
                # Move $SUBJECT_DIR to source_dir (1 time point case)
                (init_input, move_subjects_dir, [("subjects_dir", "subjects_dir")]),
                (recon_all, move_subjects_dir, [("subject_id", "subject_id")]),
                # Output node
                (move_subjects_dir, self.output_node, [("subject_id", "image_id")]),
            ]
        )
        # fmt: on



# coding: utf8

"""The 'clinica' executable command line, installed with the clinica package, calls this module.

The aim of this module is to execute pipelines from the command line,
and gives to the user some other utils to work with the pipelines.
"""
import sys

import click

from clinica.engine.template import cli as generate_cli
from clinica.iotools.converters.cli import cli as convert_cli
from clinica.iotools.utils.cli import cli as iotools_cli
from clinica.pipelines.cli import cli as run_cli
from clinica.utils.exceptions import ClinicaException
from clinica.utils.stream import cprint

CONTEXT_SETTINGS = dict(
    # Extend content width to avoid shortening of pipeline help.
    max_content_width=160,
    # Display help string with -h, in addition to --help.
    help_option_names=["-h", "--help"],
)


def setup_logging(verbosity: int = 0) -> None:
    """Setup Clinica's logging facilities.

    Args:
        verbosity (int): The desired level of verbosity for logging.
            (0 (default): WARNING, 1: INFO, 2: DEBUG)
    """
    from logging import getLogger
    from sys import stdout

    import nipype
    from colorlog import ColoredFormatter, StreamHandler

    # Cap max verbosity level to 2.
    verbosity = min(verbosity, 2)

    # Define the module level logger.
    logger = getLogger("clinica")
    logging_level = ["WARNING", "INFO", "DEBUG"][verbosity]
    logger.setLevel(logging_level)

    # Add console handler with custom formatting.
    console_handler = StreamHandler(stdout)
    console_handler.setFormatter(
        ColoredFormatter("%(log_color)s%(asctime)s:%(levelname)s:%(message)s")
    )
    logger.addHandler(console_handler)

    # Normalize nipype logging with Clinica's.
    nipype.config.set("logging", "interface_level", logging_level)
    nipype.config.set("logging", "workflow_level", logging_level)
    nipype.logging.update_logging(nipype.config)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option()
@click.option(
    "-v", "--verbose", "verbosity", count=True, help="Increase logging verbosity."
)
def cli(verbosity: int = 0) -> None:
    setup_logging(verbosity=verbosity)


cli.add_command(convert_cli)
cli.add_command(generate_cli)
cli.add_command(iotools_cli)
cli.add_command(run_cli)


def main() -> None:
    try:
        cli()
    except ClinicaException as e:
        cprint(msg=e, lvl="error")
        sys.exit(1)


if __name__ == "__main__":
    main()



# coding: utf8

"""
    This file contains a set of functional tests designed to check the correct execution of the pipeline and the
    different functions available in Clinica
"""

import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--input_data_directory",
        action="store",
        help="Directory for (only-read) inputs for tests",
    )
    parser.addoption(
        "--working_directory", action="store", help="Working directory for tests"
    )


@pytest.fixture
def cmdopt(request):
    config_param = {}
    config_param["input"] = request.config.getoption("--input_data_directory")
    config_param["wd"] = request.config.getoption("--working_directory")
    return config_param



import pathlib
from os.path import abspath, dirname, join

import pkg_resources
from setuptools import find_packages, setup

with pathlib.Path("requirements.txt").open() as requirements_txt:
    install_requires = [
        str(requirement)
        for requirement in pkg_resources.parse_requirements(requirements_txt)
    ]

with open(join(dirname(__file__), "clinica/VERSION"), "rb") as f:
    version = f.read().decode("ascii").strip()

this_directory = abspath(dirname(__file__))
with open(join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="clinica",
    version=version,
    url="http://www.clinica.run",
    description="Software platform for clinical neuroimaging studies",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="ARAMIS Lab",
    maintainer="Clinica developers",
    maintainer_email="clinica-user@inria.fr",
    license="MIT license",
    packages=find_packages(exclude=("tests", "tests.*")),
    include_package_data=True,
    zip_safe=False,
    entry_points={"console_scripts": ["clinica = clinica.cmdline:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
    ],
    install_requires=install_requires,
    extras_require={"test": ["pytest", "coverage"]},
    python_requires=">=3.6,<3.8",
)



# coding: utf8


"""This module contains utilities for PET data handling."""


def read_psf_information(pvc_psf_tsv, subject_ids, session_ids, pet_tracer):
    """Read PSF information from TSV file.

    Args:
        pvc_psf_tsv: TSV file containing participant_id, session_id, acq_label,
            psf_x, psf_y & psf_z columns
        subject_ids: list of participant IDs (e.g. ['sub-CLNC01', 'sub-CLNC01'])
        session_ids: list of session IDs (e.g. ['ses-M00', 'ses-M18'])
        pet_tracer: Tracer we want to select in acq_label column. Other tracers
            will not be read in this function

    Example of pvc_psf_tsv:

    participant_id    session_id     acq_label     psf_x    psf_y    psf_z
    sub-CLNC01        ses-M00        FDG           8        9        10
    sub-CLNC01        ses-M18        FDG           8        9        10
    sub-CLNC01        ses-M00        AV45          7        6        5
    sub-CLNC02        ses-M00        FDG           8        9        10
    sub-CLNC03        ses-M00        FDG           8        9        10

    Returns:
        PSF information following [subject_ids, session_ids] order
    """
    import os

    from pandas.io.parsers import read_csv

    if not os.path.isfile(pvc_psf_tsv):
        raise FileNotFoundError(f"Could not find the psf_tsv file {pvc_psf_tsv}")
    try:
        psf_df = read_csv(pvc_psf_tsv, sep="\t")
    except (IOError, UnicodeDecodeError):
        raise RuntimeError("An error while reading {pvc_psf_tsv} happened")

    if any(
        elem
        not in ["participant_id", "session_id", "acq_label", "psf_x", "psf_y", "psf_z"]
        for elem in list(psf_df.columns)
    ):
        raise IOError(
            f"The file {pvc_psf_tsv} must contain the following columns (separated by tabulations):\n"
            f"participant_id, session_id, acq_label, psf_x, psf_y, psf_z\n"
            f"{str(list(psf_df.columns))}\n"
            f"Pay attention to the spaces (there should be none)."
        )

    subjects_psf = list(psf_df.participant_id)
    sessions_psf = list(psf_df.session_id)
    pet_tracer_psf = list(psf_df.acq_label)
    idx_reordered = []
    for i, sub in enumerate(subject_ids):
        current_ses = session_ids[i]
        idx_sub = [
            j
            for j in range(len(subjects_psf))
            if (sub == subjects_psf[j])
            and (current_ses == sessions_psf[j])
            and (pet_tracer == pet_tracer_psf[j])
        ]
        if len(idx_sub) == 0:
            raise RuntimeError(
                f"Subject {sub} with session {current_ses} and tracer {pet_tracer} "
                f"that you want to proceed was not found in the TSV file containing "
                f"PSF specifications ({pvc_psf_tsv})."
            )
        if len(idx_sub) > 1:
            raise RuntimeError(
                f"Subject {sub} with session {current_ses} and tracer {pet_tracer} "
                f"that you want to proceed was found multiple times "
                f"in the TSV file containing PSF specifications ({pvc_psf_tsv})."
            )
        idx_reordered.append(idx_sub[0])

    psf_x = list(psf_df.psf_x)
    psf_y = list(psf_df.psf_y)
    psf_z = list(psf_df.psf_z)
    iterables_psf = [[psf_x[i], psf_y[i], psf_z[i]] for i in idx_reordered]
    return iterables_psf


LIST_SUVR_REFERENCE_REGIONS = ["pons", "cerebellumPons", "pons2", "cerebellumPons2"]


def get_suvr_mask(suvr_reference_region):
    """Get path of the SUVR mask from SUVR reference region label.

    Args:
        suvr_reference_region: Label of the SUVR reference region

    Returns:
        Path of the SUVR mask
    """
    import os

    suvr_reference_region_to_suvr = {
        "pons": os.path.join(
            os.path.split(os.path.realpath(__file__))[0],
            "..",
            "resources",
            "masks",
            "region-pons_eroded-6mm_mask.nii.gz",
        ),
        "cerebellumPons": os.path.join(
            os.path.split(os.path.realpath(__file__))[0],
            "..",
            "resources",
            "masks",
            "region-cerebellumPons_eroded-6mm_mask.nii.gz",
        ),
        "pons2": os.path.join(
            os.path.split(os.path.realpath(__file__))[0],
            "..",
            "resources",
            "masks",
            "region-pons_remove-extrabrain_eroded-2it_mask.nii.gz",
        ),
        "cerebellumPons2": os.path.join(
            os.path.split(os.path.realpath(__file__))[0],
            "..",
            "resources",
            "masks",
            "region-cerebellumPons_remove-extrabrain_eroded-3it_mask.nii.gz",
        ),
    }
    return suvr_reference_region_to_suvr[suvr_reference_region]



# coding: utf8


"""This module contains utilities to check dependencies before running Clinica.

These functions can check binaries, software (e.g. FreeSurfer) or toolboxes (e.g. SPM).
"""


def is_binary_present(binary):
    """Check if a binary is present.

    This function checks if the program is present. Do not use this function
    with a binary GUI, it will open the GUI.

    Taken from:
    https://stackoverflow.com/questions/11210104/check-if-a-program-exists-from-a-python-script

    Args:
        binary (str): Name of the program.

    Returns:
        True if the binary is present, False otherwise.
    """
    import os
    import subprocess

    from clinica.compat import errno

    try:
        devnull = open(os.devnull)
        subprocess.Popen([binary], stdout=devnull, stderr=devnull).communicate()
    except OSError as e:
        if e.errno == errno.ENOENT:
            return False
    return True


def check_environment_variable(environment_variable, software_name):
    import os

    from .exceptions import ClinicaMissingDependencyError

    content_var = os.environ.get(environment_variable, "")
    if not content_var:
        raise ClinicaMissingDependencyError(
            f"Clinica could not find {software_name} software: "
            f"the {environment_variable} variable is not set."
        )
    if not os.path.isdir(content_var):
        raise ClinicaMissingDependencyError(
            f"The {environment_variable} environment variable "
            f"you gave is not a folder (content: {content_var})."
        )
    return content_var


def check_software_requirements(current_version, version_requirements, software_name):
    from string import punctuation

    from clinica.utils.exceptions import ClinicaMissingDependencyError

    comparison_operator = "".join(
        [c for c in version_requirements if c in punctuation.replace(".", "")]
    )
    required_version = version_requirements.replace(comparison_operator, "")

    satisfy_version = eval(
        f"LooseVersion('{current_version}') {comparison_operator} LooseVersion('{required_version}')"
    )

    if not satisfy_version:
        raise ClinicaMissingDependencyError(
            f"Your {software_name} version ({current_version}) "
            f"does not satisfy version requirements ({version_requirements})."
        )


def check_dcm2niix():
    """Check dcm2niix software."""
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    if not is_binary_present("dcm2niix"):
        raise ClinicaMissingDependencyError(
            "Clinica could not find dcm2niix software in your PATH environment: "
            "this can be downloaded or installed from https://github.com/rordenlab/dcm2niix."
        )


def check_ants(version_requirements=None):
    """Check ANTs software."""
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    check_environment_variable("ANTSPATH", "ANTs")

    list_binaries = ["N4BiasFieldCorrection", "antsRegistrationSyNQuick.sh"]
    for binary in list_binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                "Clinica could not find ANTs software: "
                f"the {binary} command is not present in your PATH environment."
            )


def check_freesurfer(version_requirements=None):
    """Check FreeSurfer software."""
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    check_environment_variable("FREESURFER_HOME", "FreeSurfer")

    list_binaries = ["mri_convert", "recon-all"]
    for binary in list_binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                "Clinica could not find FreeSurfer software: "
                f"the {binary} command is not present in your PATH environment: "
                "did you have the line `source $FREESURFER_HOME/SetUpFreeSurfer.sh` "
                "in your configuration file?"
            )


def check_fsl(version_requirements=None):
    """Check FSL software."""
    import nipype.interfaces.fsl as fsl

    from clinica.utils.exceptions import ClinicaMissingDependencyError
    from clinica.utils.stream import cprint

    check_environment_variable("FSLDIR", "FSL")

    try:
        if fsl.Info.version().split(".") < ["5", "0", "5"]:
            raise ClinicaMissingDependencyError(
                "FSL version must be greater than 5.0.5"
            )
    except Exception as e:
        cprint(msg=str(e), lvl="error")

    list_binaries = ["bet", "flirt", "fast", "first"]
    for binary in list_binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                "Clinica could not find FSL software: "
                f"the {binary} command is not present in your PATH environment."
            )


def check_mrtrix(version_requirements=None):
    """Check MRtrix software."""
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    list_binaries = ["transformconvert", "mrtransform", "dwi2response", "tckgen"]
    for binary in list_binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                f"Clinica could not find MRtrix software: "
                f"the {binary} command is not present in your PATH environment."
            )


def check_petpvc(version_requirements=None):
    """Check PETPVC software."""
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    list_binaries = [
        "petpvc",
        "pvc_diy",
        "pvc_gtm",
        "pvc_iy",
        "pvc_labbe",
        "pvc_make4d",
        "pvc_mg",
        "pvc_mtc",
        "pvc_rbv",
        "pvc_relabel",
        "pvc_rl",
        "pvc_simulate",
        "pvc_stc",
        "pvc_vc",
    ]
    for binary in list_binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                "Clinica could not find PETPVC software: "
                f"the {binary} command is not present in your PATH environment."
            )


def check_spm(version_requirements=None):
    """Check SPM software."""
    check_environment_variable("SPM_HOME", "SPM")

    # list_binaries = ['matlab']
    # for binary in list_binaries:
    #     if not is_binary_present(binary):
    #         raise RuntimeError(
    #             f"{binary} from SPM Software is not present in your PATH environment.
    #           )


def check_matlab():
    """Check Matlab toolbox."""
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    if not is_binary_present("matlab"):
        raise ClinicaMissingDependencyError(
            "Matlab was not found in PATH environment. Did you add it?"
        )



# coding: utf-8

"""Module for converting FLAIR of ADNI."""


def convert_adni_flair(
    source_dir, csv_dir, dest_dir, conversion_dir, subjs_list=None, mod_to_update=False
):
    """Convert FLAIR images of ADNI into BIDS format.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        conversion_dir: path to the TSV files including the paths to original images
        subjs_list: subjects list
        mod_to_update: If True, pre-existing images in the BIDS directory will be erased and extracted again.
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids
    from clinica.utils.stream import cprint

    if not subjs_list:
        adni_merge_path = path.join(csv_dir, "ADNIMERGE.csv")
        adni_merge = pd.read_csv(adni_merge_path, sep=",", low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint(
        f"Calculating paths of FLAIR images. Output will be stored in {conversion_dir}."
    )
    images = compute_flair_paths(
        source_dir, csv_dir, dest_dir, subjs_list, conversion_dir
    )
    cprint("Paths of FLAIR images found. Exporting images into BIDS ...")
    # flair_paths_to_bids(images, dest_dir)
    paths_to_bids(images, dest_dir, "flair", mod_to_update=mod_to_update)
    cprint(msg="FLAIR conversion done.", lvl="debug")


def compute_flair_paths(source_dir, csv_dir, dest_dir, subjs_list, conversion_dir):
    """Compute the paths to the FLAIR images and store them in a TSV file.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list
        conversion_dir: path to the TSV files including the paths to original images

    Returns:
        images: a dataframe with all the paths to the FLAIR images that will be converted into BIDS
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        find_image_path,
        visits_to_timepoints,
    )

    flair_col_df = [
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Series_ID",
        "Image_ID",
        "Field_Strength",
        "Scanner",
    ]
    flair_df = pd.DataFrame(columns=flair_col_df)
    flair_dfs_list = []

    # Loading needed .csv files
    adni_merge = pd.read_csv(
        path.join(csv_dir, "ADNIMERGE.csv"), sep=",", low_memory=False
    )

    mayo_mri_qc = pd.read_csv(
        path.join(csv_dir, "MAYOADIRL_MRI_IMAGEQC_12_08_15.csv"),
        sep=",",
        low_memory=False,
    )
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == "AFL"]

    mri_list = pd.read_csv(path.join(csv_dir, "MRILIST.csv"), sep=",", low_memory=False)

    # Selecting FLAIR DTI images that are not MPR
    mri_list = mri_list[mri_list.SEQUENCE.str.contains("flair", case=False, na=False)]
    unwanted_sequences = ["_MPR_"]
    mri_list = mri_list[
        mri_list.SEQUENCE.map(
            lambda x: not any(subs in x for subs in unwanted_sequences)
        )
    ]

    for subj in subjs_list:

        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values("EXAMDATE")

        mri_list_subj = mri_list[mri_list.SUBJECT == subj]
        mri_list_subj = mri_list_subj.sort_values("SCANDATE")

        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(subj, mri_list_subj, adnimerge_subj, "FLAIR")

        for visit_info in visits.keys():
            timepoint = visit_info[0]
            visit_str = visits[visit_info]

            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]
            flair = flair_image(
                subj, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj
            )

            if flair is not None:
                row_to_append = pd.DataFrame(
                    flair,
                    index=[
                        "i",
                    ],
                )
                flair_dfs_list.append(row_to_append)

    if flair_dfs_list:
        flair_df = pd.concat(flair_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # Eq_1 images
        ("141_S_0767", "m84"),
        ("067_S_5205", "bl"),
        ("127_S_4928", "m24"),
        ("024_S_4674", "m06"),
        ("123_S_2363", "m24"),
        ("053_S_4578", "m48"),
        ("128_S_4586", "m48"),
        ("053_S_4813", "m48"),
        ("053_S_5272", "m24"),
        ("013_S_1186", "m48"),
        ("031_S_2022", "bl"),
        ("031_S_2022", "m06"),
        ("031_S_2233", "bl"),
        ("031_S_2233", "m03"),
        ("029_S_2395", "m72"),
        ("130_S_6043", "bl"),
        ("031_S_2018", "bl"),
        ("027_S_5170", "m72"),
        ("135_S_6284", "m12"),
        ("068_S_0127", "m180"),
        ("068_S_2187", "m120"),
        # Several output images
        ("114_S_6039", "bl"),
    ]

    # Removing known exceptions from images to convert
    if not flair_df.empty:
        error_ind = flair_df.index[
            flair_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors), axis=1
            )
        ]
        flair_df.drop(error_ind, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(flair_df, source_dir, "FLAIR", "S", "Series_ID")
    images.to_csv(path.join(conversion_dir, "flair_paths.tsv"), sep="\t", index=False)

    return images


def flair_image(subject_id, timepoint, visit_str, visit_mri_list, mri_qc_subj):
    """
    One image among those in the input list is chosen according to QC
    and then correspoding metadata is extracted to a dictionary

    Args:
        subject_id: Subject identifier
        timepoint: Visit code
        visit_str: Visit name
        visit_mri_list: List of images metadata
        mri_qc_subj: Dataframe containing list of QC of scans for the subject

    Returns: dictionary - contains image metadata
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        replace_sequence_chars,
        select_image_qc,
    )

    sel_image = select_image_qc(list(visit_mri_list.IMAGEUID), mri_qc_subj)
    if sel_image is None:
        return None

    sel_scan = visit_mri_list[visit_mri_list.IMAGEUID == sel_image].iloc[0]

    image_dict = {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": replace_sequence_chars(sel_scan.SEQUENCE),
        "Scan_Date": sel_scan["SCANDATE"],
        "Study_ID": str(int(sel_scan.STUDYID)),
        "Series_ID": str(int(sel_scan.SERIESID)),
        "Image_ID": str(int(sel_scan.IMAGEUID)),
        "Field_Strength": sel_scan.MAGSTRENGTH,
    }

    return image_dict
